/* Miscellaneous functions

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/services.h"

static FILE *log_file;
static FILE *debug_file;
static int ns_iterator_pos = 1024;
static NickInfo *ns_iterator_ptr = NULL;
static int cs_iterator_pos = 1024;
static ChanInfo *cs_iterator_ptr = NULL;

/* Open the logfile */
void open_log ()
{
    if (log_file)
	return;

    log_file = fopen (LOGFILE, "a");

    if (log_file)
	setbuf (log_file, NULL);
}

/* Open the debug log */
void open_debug ()
{
    if (debug_file)
	return;

    debug_file = fopen ("debug.log", "a");

    if (debug_file)
	setbuf (debug_file, NULL);
}

/* Log something. */
void log (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[32];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);

    if (!log_file)
	open_log ();

    if (log_file)
    {
	fputs (buf, log_file);
	vfprintf (log_file, fmt, args);
	fputc ('\n', log_file);
	fflush (log_file);
    }

    if (runflags & RUN_LIVE)
    {
	fputs (buf, stderr);
	vfprintf (stderr, fmt, args);
	fputc ('\n', stderr);
    }
}

/* Special log handling. If sameline is 1, the
   entry will have a timestamp, but no newline.
   Otherwise it'll have a newline, but no timestamp.
 */
void log_sameline (int sameline, const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[32];

    va_start (args, fmt);

    if (sameline)
    {
	time (&t);
	tm = *localtime (&t);
	strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);
    }

    if (!log_file)
	open_log ();

    if (log_file)
    {
	if (sameline)
	    fputs (buf, log_file);

	vfprintf (log_file, fmt, args);

	if (!sameline)
	    fputc ('\n', log_file);

	fflush (log_file);
    }

    if (runflags & RUN_LIVE)
    {
	if (sameline)
	    fputs (buf, stderr);

	vfprintf (stderr, fmt, args);

	if (!sameline)
	    fputc ('\n', stderr);
    }
}

/* Write to our debug log */
void debug (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[32];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);

    if (!debug_file)
	open_debug ();

    if (debug_file)
    {
	fputs (buf, debug_file);
	vfprintf (debug_file, fmt, args);
	fputc ('\n', debug_file);
	fflush (debug_file);
    }

    if (runflags & RUN_LIVE)
    {
	fputs (buf, stderr);
	vfprintf (stderr, fmt, args);
	fputc ('\n', stderr);
    }
}

/* Close the logfile */
void close_log ()
{
    if (!log_file)
	return;

    fclose (log_file);

    log_file = NULL;
}

/* Close the debug log */
void close_debug ()
{
    if (!debug_file)
	return;

    fclose (debug_file);

    debug_file = NULL;
}

/* Fatal error: Log it, globop it if we're still connected, then die. */
void fatal (const char *fmt,...)
{
    va_list args;
    char buf[BUFSIZE];

    va_start (args, fmt);    
    vsnprintf (buf, sizeof (buf), fmt, args);

    log ("FATAL: %s", buf);

    if (servsock > 0)
	globops ("Fatal Error: %s", buf);

    exit (1);
}

/* Send a command to the server.  The two forms here are like
   printf()/vprintf() and friends.
 */
void send_cmd (const char *source, const char *fmt,...)
{
    va_list args;
   
    va_start (args, fmt);
    vsend_cmd (source, fmt, args);
    va_end (args);
}

void vsend_cmd (const char *source, const char *fmt, va_list args)
{
    char buf[2048];
    
    vsnprintf (buf, sizeof (buf), fmt, args);

    if (source)
    {
	sendtoserver (":%s %s\r\n", source, buf);
	if (debuglevel == 1 || debuglevel == 3)
	    debug ("<-- %s %s", source, buf);
    }
    else
    {
	sendtoserver ("%s\r\n", buf);
	if (debuglevel == 1 || debuglevel == 3)
	    debug ("<-- %s", buf);
    }
}

/* Send a globops */
void globops (const char *sender, const char *fmt,...)
{
    va_list args;
    char buf[2048];
     
    va_start (args, fmt);
#ifdef IRCUP9
    /* IRCu doesn't have GlobOps. We'll use wallops. */
    snprintf (buf, sizeof (buf), "WALLOPS :%s", fmt);
#else
    snprintf (buf, sizeof (buf), "%s :%s", me.token ? "]" : "GLOBOPS", fmt);
#endif

    vsend_cmd (sender ? sender : me.name, buf, args);
    va_end (args);
}

/* PRIVMSG our snooping channel */
void snoop (const char *source, const char *fmt,...) 
{
    va_list args;
    char buf[2048];
    time_t t;
    struct tm tm;
    char timebuf[256];

    if (findchan (snoopchan))
    {
	time (&t);
	tm = *localtime (&t);
	strftime (timebuf, sizeof (timebuf) - 1, "[%H:%M:%S]", &tm);
	va_start (args, fmt);
	snprintf (buf, sizeof (buf), "%s %s :%s %s", me.token ? "!" :
	    "PRIVMSG", snoopchan, timebuf, fmt);
	vsend_cmd (source, buf, args);
	va_end (args);
    }
}

/* Send a notice to a user */
void notice (const char *source, const char *dest, const char *fmt,...)
{
    NickInfo *ni;
    User *u = finduser (dest);
    va_list args;
    char buf[2048];

    if (!u)
	return;

    ni = get_nick (u);

    va_start (args, fmt);

    if ((privmsg_on == FALSE && ni && (ni->flags & NF_PRIVMSG)) ||
	(privmsg_on == TRUE && ((ni && !(ni->flags & NF_NOTICE)) ||
				!ni)))
	snprintf (buf, sizeof (buf), "PRIVMSG %s :%s", dest, fmt);
    else
	snprintf (buf, sizeof (buf), "NOTICE %s :%s", dest, fmt);

    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Send an op notice to a channel */
void opnotice (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "NOTICE @%s :%s", dest, fmt);
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Reply to a CTCP command. */
void ctcpreply (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "%s %s :\001%s\001", me.token ? "B" :
	"NOTICE", dest, fmt);
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Send a global notice to all users. */
void noticeall (const char *fmt,...)
{
#ifndef UNREAL3
    char *tmp = strtok (tlds, " ");
#endif

    /* Is the GlobalNoticer on? Bail out if not.
       (We could easily use RootServ for this, but if people don't have
       the GlobalNoticer on, they probably use something else for notices.)
     */
    if (globalnoticer_on == FALSE)
	return;

#ifdef UNREAL3
    /* Unreal lets us notice $*, so we'll do that. */
    send_cmd (s_GlobalNoticer, "%s $* :%s", me.token ? "B" : "NOTICE", fmt);
#else
    
    /* Notice the first TLD */
    send_cmd (s_GlobalNoticer, "%s $*.%s :%s", me.token ? "B" : "NOTICE",
	tmp, fmt);
        
    /* Now run through the list and notice the remaining tlds.
       This will terminate when we run out of tlds..
     */
    for (;;)
    {
	tmp = strtok (NULL, " ");

	if (tmp)
	    send_cmd (s_GlobalNoticer, "%s $*.%s :%s", me.token ? "B" :
		"NOTICE", tmp, fmt);
	else
	    break;
    }
#endif /* UNREAL3 */
}       

/* Case insensitive version of strcmp */
int stricmp (const char *s1, const char *s2)
{
    register int c;

    while ((c = tolower (*s1)) == tolower (*s2))
    {
	if (c == 0)
	    return 0;
	s1++;
	s2++;
    }
    if (c < tolower (*s2))
	return -1; 
    return 1;
}

/* strscpy:  Copy at most len-1 characters from a string to a buffer, and
             add a null terminator after the last character copied.
 */
char *strscpy (char *d, const char *s, size_t len)
{
    char *d_orig = d;

    if (!len)
	return d;
    while (--len && (*d++ = *s++))
	;
    *d = 0;
    return d_orig;
}

void *smalloc (long size)
{
    void *buf;

    buf = malloc (size);
    if (!buf)
        raise (SIGUSR1);
    return buf;
}  

void *scalloc (long elsize, long els)  
{
    void *buf = calloc (elsize, els);
    if (!buf)
	raise (SIGUSR1);
    return buf;
}

void *srealloc (void *oldptr, long newsize)
{   
    void *buf = realloc (oldptr, newsize);
    if (!buf)
        raise (SIGUSR1);
    return buf;
}

char *sstrdup (const char *s)
{   
    char *t = strdup (s);
    if (!t)
        raise (SIGUSR1);
    return t;
}

/* strlcat: Not all systems (ie, linux) have strlcat, so here it is. */
size_t sstrlcat (char *dst, const char *src, size_t siz)
{
    char *d = dst;
    const char *s = src;
    size_t n = siz, dlen;

    while (*d != '\0' && n-- != 0)
	d++;

    dlen = d - dst;
    n = siz - dlen;

    if (!n)
	return (dlen + strlen (s));

    while (*s != '\0')
    {
	if (n != 1)
	{
	    *d++ = *s;
	    n--;
	}

	s++;
    }

    *d = '\0';

    return (dlen + (s - src));
}

/* strlcpy: Not all systems (ie, linux) have strlcpy, so here it is. */
size_t sstrlcpy (char *dst, const char *src, size_t siz)
{
    char *d = dst;
    const char *s = src;
    size_t n = siz;

    if (n != 0 && --n != 0)
    {
	do
	{
	    if (!(*d++ = *s++))
		break;
	} while (--n != 0);
    }

    if (!n)
    {
	if (siz != 0)
	    *d = '\0';

	while (*s++);
    }

    return (s - src - 1);
}

/* Hash function for commands */
Hash *get_hash (const char *whoami, const char *source, const char *cmd,
		Hash * hash_table)
{
    for (; hash_table->accept; ++hash_table)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_CSOP && is_csop (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_CSOP && !is_csop (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}

    notice (whoami, source, ERR_UNKNOWN_COMMAND, cmd, haveserv_on == TRUE ? "" : "MSG ",
	whoami, haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");

    return NULL;
}

/* Subcommand hash - Similar to above, but no error if the command isn't
   found, the function that called this will handle that.
 */
Hash_SUB *get_sub_hash (const char *whoami, const char *source, const char *cmd,
			Hash_SUB * hash_table)
{
    for (; hash_table->accept; ++hash_table)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_CSOP && is_csop (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_CSOP && !is_csop (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}
    return NULL;
}

/* Hash function for help tables */
Hash_HELP *get_help_hash (const char *whoami, const char *source,
			  const char *cmd, Hash_HELP * hash_table)
{
    for (; hash_table->accept; ++hash_table)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_CSOP && is_csop (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_CSOP && !is_csop (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}
    notice (whoami, source, ERR_NO_HELP, cmd);
    return NULL;
}

/* Is the given nick an oper? */
int is_oper (const char *nick)  
{
    User *user = finduser (nick);

    return (user && (user->mode & UMODE_o));
}

/* Is the given nick on the given channel? */
int is_on_chan (const char *nick, const char *chan)
{
    Channel *c = findchan (chan);
    struct c_userlist *cu;

    if (!c)
	return 0;

    for (cu = c->users; cu; cu = cu->next)
	if (!stricmp (cu->user->nick, nick))
	   return 1;

    return 0;
}

/* Is the given nick a chanop on the given channel? */
int is_opped (const char *nick, const char *chan)
{
    Channel *c = findchan (chan);
    struct c_userlist *cu;

    if (!c)
	return 0;

    for (cu = c->users; cu; cu = cu->next)
	if (!stricmp (cu->user->nick, nick))
	    break;

    return (cu && cu->mode & CMODE_o);
}

/* Is the given nick a voice on the given channel? */
int is_voiced (const char *nick, const char *chan)
{
    Channel *c = findchan (chan);
    struct c_userlist *cu;

    if (!c)
	return 0;

    for (cu = c->users; cu; cu = cu->next)
	if (!stricmp (cu->user->nick, nick))
	    break;

    return (cu && cu->mode & CMODE_v);
}

#ifdef UNREAL3
/* Is the given nick a halfop on the given channel? */
int is_halfopped (const char *nick, const char *chan)
{
    Channel *c = findchan (chan);
    struct c_userlist *cu;

    if (!c)
	return 0;

    for (cu = c->users; cu; cu = cu->next)
	if (!stricmp (cu->user->nick, nick))
	    break;

    return (cu && cu->mode & CMODE_h);
}
#endif

/* Is the given nick a CSOp? */
int is_csop (const char *nick)
{
    NickInfo *ni;
    User *u = finduser (nick);

    /* SRA > CSOp :P */
    if (is_sra (nick))
	return 1;

    ni = get_nick (u);

    if (!ni)
	return 0;

    if (ni->flags & NF_CSOP)
	return 1;

    return 0;
}

/* Could this nick become a CSOp? */
int can_csop (const char *nick)
{
    NickInfo *ni = findnick (nick);

    /* SRA > CSOp .. :P */
    if (can_sra (nick))
	return 1;

    if (!ni)
	return 0;

    if (ni->flags & NF_CSOP)
	return 1;

    return 0;
}

/* Is the given nick an SRA? */
int is_sra (const char *nick)
{
    User *user = finduser (nick);

    return (user && (user->flags & UF_SRA));
}

/* Could this nick become an SRA? */
int can_sra (const char *nick)
{
    if (findsra (nick))
	return 1;

    return 0;
}

/* Search case-insensitively for string s2 within string s1 */
char *stristr (char *s1, char *s2)
{ 
    register char *s = s1, *d = s2;

    while (*s1)
	if (tolower (*s1) == tolower (*d))
	{
	    s1++;
	    d++; 
	    if (*d == 0)
		return s;
	}
	else
	{
	    s = ++s1;
	    d = s2;  
	}
    return NULL;
}

/* Find a wildcard match */
static int do_match_wild (const char *pattern, const char *str, int  docase)
{
    char c;
    const char *s;

    for (;;)
    {
	switch (c = *pattern++)
	{
	    case 0:
		if (!*str)
		    return 1;
		return 0;
	    case '?':
		if (!*str)
		    return 0;
		str++;
		break;
	    case '*':
		if (!*pattern)
		    return 1;
		s = str;
		while (*s)
		{
		    if ((docase ? (*s==*pattern) : (tolower (*s)==tolower
			(*pattern))) && do_match_wild (pattern, s, docase))
			return 1;
		    s++;
		}
		break;
	    default:
		if (docase ? (*str++ != c) : (tolower(*str++) != tolower(c)))
		    return 0;
		break;
	}
    }
}

/* Call do_match_wild case-sensitively */
int match_wild (const char *pattern, const char *str)
{
    return do_match_wild (pattern, str, 1);
}

/* Call do_match_wild case-insensitively */
int match_wild_nocase (const char *pattern, const char *str)
{
    return do_match_wild (pattern, str, 0);
}

/* strupper: Convert a string to upper case. */
char *strupper (char *s)
{
    char *t = s;
    while (*t)
	*t++ = toupper (*t);
    return s;
}

/* strlower: Convert a string to lower case. */
char *strlower (char *s)
{
    char *t = s;
    while (*t)
	*t++ = tolower (*t);
    return s;
}

/* Strip unwanted chars out of lines - Borrowed from GeoStats */
void strip (char *line)
{  
    char *c;

    if (line)
    {
	if ((c = strchr (line, '\n')))
	    *c = '\0';
	if ((c = strchr (line, '\r')))
	    *c = '\0';
	if ((c = strchr (line, '\1')))
	    *c = '\0';
    }
}

/* Is it one of ours? */
int is_one_of_mine (const char *nick)
{
    if ((rootserv_on == TRUE && !stricmp (nick, s_RootServ)) ||
        (nickserv_on == TRUE && !stricmp (nick, s_NickServ)) ||
	(chanserv_on == TRUE && !stricmp (nick, s_ChanServ)) ||
	(memoserv_on == TRUE && !stricmp (nick, s_MemoServ)) ||
	(globalnoticer_on == TRUE && !stricmp (nick, s_GlobalNoticer)))
        return 1;
 
    return 0;
}

/* Return the date an event happened
   If fulldate is:
	0:	Sun, Jan 01, 2000
	1:	Sun, Jan 01, 2000, 00:00:00 GMT
	2:	00:00:00 GMT
	3:	01/01/2000, 00:00:00 GMT
	4:	Jan 01, 2000, 00:00:00 GMT
 */
char *get_time (time_t event, int fulldate)
{
    struct tm tm;
    static char timebuf[256];

    tm = *gmtime (&event);

    /* Date and time, or just date */
    if (fulldate == 1)
	strftime (timebuf, sizeof (timebuf), "%a, %b %d, %Y, %H:%M:%S GMT",
	    &tm);
    else if (fulldate == 2)
	strftime (timebuf, sizeof (timebuf), "%H:%M:%S GMT", &tm);
    else if (fulldate == 3)
	strftime (timebuf, sizeof (timebuf), "%d/%m/%Y %H:%M:%S GMT", &tm);
    else if (fulldate == 4)
	strftime (timebuf, sizeof (timebuf), "%b %d, %Y, %H:%M:%S GMT", &tm);
    else
	strftime (timebuf, sizeof (timebuf), "%a, %b %d, %Y", &tm);

    timebuf[sizeof(timebuf)-1] = 0;
    return timebuf;
}

/* Return the time elapsed since an event */
char *time_ago (time_t event)
{
    static char ret[128];
    int years, weeks, days, hours, minutes, seconds;

    event = time (NULL) - event;
    years = weeks = days = hours = minutes = seconds = 0;

    while (event >= 60 * 60 * 24 * 365)
    {
	event -= 60 * 60 * 24 * 365;
	years++;
    }
    while (event >= 60 * 60 * 24 * 7)
    {
	event -= 60 * 60 * 24 * 7;
	weeks++;
    }
    while (event >= 60 * 60 * 24)
    {
	event -= 60 * 60 * 24;
	days++; 
    }
    while (event >= 60 * 60)
    {
	event -= 60 * 60;
	hours++;
    }
    while (event >= 60)
    {
	event -= 60;
	minutes++;
    }

    seconds = event;
     
    if (years)
	snprintf (ret, sizeof (ret),
		"%d year%s, %d week%s, %d day%s, %02d:%02d:%02d",
		years, years == 1 ? "" : "s",
		weeks, weeks == 1 ? "" : "s",
		days, days == 1 ? "" : "s",
		hours, minutes, seconds);
    else if (weeks)
	snprintf (ret, sizeof (ret), "%d week%s, %d day%s, %02d:%02d:%02d",
		weeks, weeks == 1 ? "" : "s",
		days, days == 1 ? "" : "s",
		hours, minutes, seconds);
    else if (days)
	snprintf (ret, sizeof (ret), "%d day%s, %02d:%02d:%02d",
		days, days == 1 ? "" : "s",  
		hours, minutes, seconds);
    else if (hours)
	snprintf (ret, sizeof (ret), "%d hour%s, %d minute%s, %d second%s",
		hours, hours == 1 ? "" : "s",
		minutes, minutes == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
    else if (minutes)
	snprintf (ret, sizeof (ret), "%d minute%s, %d second%s",
		minutes, minutes == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
    else
	snprintf (ret, sizeof (ret), "%d second%s",
		seconds, seconds == 1 ? "" : "s");

    return ret;
}

/* Duration: How long since an event happened.
   Depending on format, we return a time in one of two formats:
	1: #d#h#m#s
	2: # days, # hours, # minutes, # seconds
   Like dotime, we only do days, hours, minutes and seconds here.
 */
char *duration (time_t event, int format)
{
    static char ret[128];
    int days, hours, minutes, seconds;

    days = hours = minutes = seconds = 0;

    while (event >= 60 * 60 * 24)
    {
	event -= 60 * 60 * 24;
	days++; 
    }
    while (event >= 60 * 60)
    {
	event -= 60 * 60;
	hours++;
    }
    while (event >= 60)
    {
	event -= 60;
	minutes++;
    }

    seconds = event;

    /* This is pretty big, but it works.. */
    if (format == 1)
    {
	if (days && hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%dd%dh%dm%ds", days, hours, minutes, seconds);
	else if (days && hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%dd%dh%dm", days, hours, minutes);
	else if (days && hours)
	    snprintf (ret, sizeof (ret),
		"%dd%dh", days, hours);
	else if (days)
	    snprintf (ret, sizeof (ret),
		"%dd", days);
	else if (hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%dh%dm%ds", hours, minutes, seconds);
	else if (hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%dh%dm", hours, minutes);
	else if (hours)
	    snprintf (ret, sizeof (ret),
		"%dh", hours);
	else if (minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%dm%ds", minutes, seconds);
	else if (minutes)
	    snprintf (ret, sizeof (ret), 
		"%dm", minutes);
	else
	    snprintf (ret, sizeof (ret),
		"%ds", seconds);
    }

    if (format == 2)
    {
	if (days && hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%d day%s, %d hour%s, %d minute%s, %d second%s", days, days == 1 ? "" : "s",
		hours, hours == 1 ? "" : "s", minutes, minutes == 1 ? "" : "s", seconds,
		seconds == 1 ? "" : "s");
	else if (days && hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%d day%s, %d hour%s, %d minute%s", days, days == 1 ? "" : "s",
		hours, hours == 1 ? "" : "s", minutes, minutes == 1 ? "" : "s");
	else if (days && hours)
	    snprintf (ret, sizeof (ret),
		"%d day%s, %d hour%s", days, days == 1 ? "" : "s",
		hours, hours == 1 ? "" : "s");
	else if (days)
	    snprintf (ret, sizeof (ret),
		"%d day%s", days, days == 1 ? "" : "s");
	else if (hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%d hour%s, %d minute%s, %d second%s", hours, hours == 1 ? "" : "s",
		minutes, minutes == 1 ? "" : "s", seconds, seconds == 1 ? "" : "s");
	else if (hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%d hour%s, %d minute%s", hours, hours == 1 ? "" : "s",
		minutes, minutes == 1 ? "" : "s");
	else if (hours)
	    snprintf (ret, sizeof (ret),
		"%d hour%s", hours, hours == 1 ? "" : "s");
	else if (minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%d minute%s %d second%s", minutes, minutes  == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
	else if (minutes)
	    snprintf (ret, sizeof (ret), 
		"%d minute%s", minutes, minutes == 1 ? "" : "s");
	else
	    snprintf (ret, sizeof (ret),
		"%d second%s", seconds, seconds == 1 ? "" : "s");
    }

    return ret;
}

/* dotime:  Return the number of seconds corresponding to the given time
            string.  If the given string does not represent a valid time,
            return -1.
  
            A time string is either a plain integer (representing a number
            of seconds), or an integer followed by one of these characters:
            "s" (seconds), "m" (minutes), "h" (hours), or "d" (days).

   Borrowed from IRCServices, modified to accept stacking of times. -skold
 */
int dotime (const char *s)
{
    int amount, seconds = 0;

    amount = strtol (s, (char **)&s, 10);

    while (*s)
    {
	switch (*s++)
	{
	    case 's':
		seconds += amount;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'm':
		seconds += amount * 60;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'h':
		seconds += amount * 3600;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'd':
		seconds += amount * 86400;
		amount = strtol (s, (char **)&s, 10);
		break;
	}
    }
    return seconds;
}

/* Send a NULL-terminated array of text as NOTICEs. */
void notice_list (const char *source, const char *dest, const char **text)
{
    while (*text)
    {
	if (**text)
	    notice (source, dest, *text);
	else
	    notice (source, dest, " ");
	text++;  
    }
}

/* Find an SRA by nick. Return NULL if sra could not be found. */
SRA *findsra (const char *nick)
{
    SRA *sra;

    if (!nick)  
	return NULL;

    for (sra = sralist; sra; sra = sra->next)
	if (!stricmp (nick, sra->nick))
	    return sra;

    return NULL;
}

/* Find a nickname. Return NULL if nickname could not be found. */
NickInfo *findnick (const char *nick)
{
    NickInfo *ni;

    for (ni = nicklist[NSHASH(nick)]; ni; ni = ni->next)
	if (!stricmp (nick, ni->nick))
	    return ni;

    return NULL;
}

/* Return the first nickname on record */
NickInfo *firstni (void)
{
    ns_iterator_pos = -1;
    ns_iterator_ptr = NULL;

    return nextni ();
}

/* Return the next nickname on record */
NickInfo *nextni (void)
{
    if (ns_iterator_ptr)
	ns_iterator_ptr = ns_iterator_ptr->next;

    while (!ns_iterator_ptr && ns_iterator_pos < 1024)
    {
	ns_iterator_pos++;

	if (ns_iterator_pos < 1024)
	    ns_iterator_ptr = nicklist[ns_iterator_pos];
    }

    return ns_iterator_ptr;
}

/* Find a registered channel. Return NULL if channel could not be found. */
ChanInfo *cs_findchan (const char *chan)
{
    ChanInfo *ci;

    if (!chan)
	return NULL;

    for (ci = cs_chanlist[CSHASH(chan)]; ci; ci = ci->next)
	if (!stricmp (chan, ci->name))
	    return ci;

    return NULL;
}

/* Return the first channel */
ChanInfo *firstci (void)
{
    cs_iterator_pos = -1;
    cs_iterator_ptr = NULL;

    return nextci ();
}

/* Return the next channel */
ChanInfo *nextci (void)
{
    if (cs_iterator_ptr)
	cs_iterator_ptr = cs_iterator_ptr->next;

    while (!cs_iterator_ptr && cs_iterator_pos < 1024)
    {
	cs_iterator_pos++;

	if (cs_iterator_pos < 1024)
	    cs_iterator_ptr = cs_chanlist[cs_iterator_pos];
    }

    return cs_iterator_ptr;
}

/* Find a verification. Return NULL if not found. */
Verify *findverify (const char *name)
{
    Verify *vf;

    if (!name)
	return NULL;

    for (vf = verifylist[VHASH(name)]; vf; vf = vf->next)
	if (!stricmp (name, vf->name))
	    return vf;

    return NULL;
}

/* Find an E-Mail. Return NULL if not found. */
EMail *findemail (const char *email)
{
    EMail *em;

    if (!email)
	return NULL;

    for (em = emaillist[EHASH(email)]; em; em = em->next)
	if (!stricmp (email, em->email))
	    return em;

    return NULL;
}

/* Check how many times the given email has been used in the last 24 hours.
   If it's 5 or more, return 1. If not, incriment the count and return 0.
   If the email record doesnt exist, create it and return 0.
 */
int check_email (const char *source, const char *email)
{
    EMail *em;

    /* Check if this email has been used more than 5 times in the last
       24 hours.
     */
    em = findemail (email);

    if (em)
    {
	if (em->times >= 5)
	    return 1;
	else
	{
	    em->times++;
	    return 0;
	}
    }

    /* The E-Mail isnt in the records, create a new record. */
    if (!em)
    {
	EMail **elist;

	em = scalloc (sizeof (EMail), 1);
	em->email = sstrdup (email);
	em->since = time (NULL);
	em->times = 1;

	elist = &emaillist[EHASH(em->email)];
	em->next = *elist;
	if (*elist)
	    (*elist)->prev = em;
	*elist = em;

	return 0;
    }

    return 0;
}

/* Remove any email records older than 24 hours */
void expire_emails (void)
{
    EMail *em;
    int i, j = 0;

    if (logupdates_on == TRUE)
	log_sameline (1, "EM:SYNC: Checking expires... ");

    for (i = 0; i < 1024; i++)
    {
	for (em = emaillist[i]; em; em = em->next)
	{
	    if (time (NULL) - em->since >= 86400)
	    {
		free (em->email);

		if (em->prev)
		    em->prev->next = em->next;
		else
		    emaillist[EHASH(em->email)] = em->next;

		if (em->next)
		    em->next->prev = em->prev;

		free (em);
	    }
	}
    }

    if (logupdates_on == TRUE)
	log_sameline (0, "%d e-mails expired.", j);
}

/* Is the given mask akilled? */
int is_akilled (const char *mask)
{   
    AutoKill *akill;
    int i;
        
    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	    if (match_wild_nocase (akill->mask, mask))
		return 1;
    return 0;
}

/* Given a user, return a mask that will most likely match any address the
   user will have from that location.  For IP addresses, wildcards the
   appropriate subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*);
   for named addresses, wildcards the leftmost part of the name unless the
   name only contains two parts.  If the username begins with a ~, delete
   it.  The returned character string is malloc'd and should be free'd
   when done with.

   Borrowed from IRCServices.
 */
char *create_mask (User *u)
{
    char *mask, *s, *end, *host;

    host = u->host;

#ifdef UNREAL3
    if (fakehost_on == TRUE && u->fakehost)
	host = u->fakehost;
#endif

    end = mask = smalloc (strlen (u->user) + strlen (host) + 2);
    end += sprintf (end, "*@");

    if (strspn (host, "0123456789.") == strlen (host)
	&& (s = strchr (host, '.')) && (s = strchr(s+1, '.'))
	&& (s = strchr (s+1, '.'))  && (   !strchr(s+1, '.')))
    {
	s = sstrdup (host);
	*strrchr (s, '.') = 0;
	if (atoi (host) < 192)
	    *strrchr (s, '.') = 0;
	if (atoi (host) < 128)
	    *strrchr (s, '.') = 0;
	sprintf (end, "%s.*", s);
	free(s);
    }
    else
    {
	if ((s = strchr (host, '.')) && strchr (s+1, '.'))
	{
	    s = sstrdup (strchr (host, '.') -1);
	    *s = '*';
	}
	else
	    s = sstrdup (host);

	strcpy (end, s);
	free (s);
    }
    return mask;
}

/* Remove control codes from text. Borrowed from UnrealIRCd,
   modified to strip bold, underline and reverse.
 */
char *stripcodes (char *buffer)
{
    static char in[BUFSIZE], out[BUFSIZE];
    int i = 0, j = 0, hascomma = 0;

    /* Check if theres any codes in the text */
    if (!strchr (buffer, '\002') && !strchr (buffer, '\003') &&
	!strchr (buffer, '\026') && !strchr (buffer, '\037'))
	return buffer;

    bzero ((char *)out, sizeof (out));
    memcpy (in, buffer, BUFSIZE);

    while (in[i])
    {
	/* Remove codes */
	if (in[i] == '\002' || in[i] == '\003' ||
	    in[i] == '\026' || in[i] == '\037')
	{
	    hascomma = 0;
	    i++;

	    if (!isdigit (in[i]))
		continue;

	    while (isdigit (in[i]) || (isdigit (in[i - 1])
		   && in[i] == ',' && isdigit (in[i + 1])
		   && hascomma == 0))
	    {
		if (in[i] == ',' && hascomma == 1)
		    break;
		if (in[i] == ',' && hascomma == 0)
		    hascomma = 1;
		i++;
	    }
	    continue;
	}
	out[j] = in[i];
	i++;
	j++;
    }
    return out;
}

/* Replace oldpiece with newpiece in string. */
char *replace (char *string, char *oldpiece, char *newpiece)
{
    int str_index, newstr_index, oldpiece_index, end;
    int new_len, old_len, cpy_len;
    char *c;
    static char newstring[2048];

    if ((c = (char *) strstr (string, oldpiece)) == NULL)
	return string;

    new_len = strlen (newpiece);
    old_len = strlen (oldpiece);
    end = strlen (string) - old_len;
    oldpiece_index = c - string;

    newstr_index = 0;
    str_index = 0;

    while (str_index <= end && c != NULL)
    {
	/* Copy characters from the left of matched pattern occurence */
	cpy_len = oldpiece_index - str_index;
	strncpy (newstring + newstr_index, string + str_index, cpy_len);
	newstr_index += cpy_len;
	str_index += cpy_len;

	/* Copy replacement characters instead of matched pattern */
	strcpy (newstring + newstr_index, newpiece);
	newstr_index += new_len;
	str_index += old_len;

	/* Check for another pattern match */
	if ((c = (char *) strstr (string + str_index, oldpiece)) != NULL)
	   oldpiece_index = c - string;
    }

    /* Copy remaining characters from the right of last matched pattern */
    strcpy (newstring + newstr_index, string + str_index);

    return newstring;
}

/* Reverse of atoi (); */
char *itoa (int num)
{
    static char ret[32];
    sprintf (ret, "%d", num);
    return ret;
}

/* Clear all SRAs. */
void flushlists (void)
{
    SRA *sra;

    for (sra = sralist; sra; sra = sra->next)
    {
	free (sra->nick);
	free (sra->pass);

	if (sra->prev)
	    sra->prev->next = sra->next;
	else
	    sralist = sra->next;

	if (sra->next)
	    sra->next->prev = sra->prev;
    }

    free (sra);
}

/* Return the nickname the user last id'd for. NULL if they haven't.

   Here's how it works: When a user identifies for a nickname, that
   nickname is stored in their user info. This way, a user can change
   to any nickname they want, and identify from any nickname they want,
   they'll always be identified for that nickname until they disconnect,
   deauth, or identify for another nick.

   When we want to know if someones identified, we come here. If the
   nickname in their u->lastnick is a registered nickname, they're
   considered to be identified for it. It might not always be a registered
   nickname though, since it might get dropped by someone else who
   identified for it.

   Users are always considered to be wanting the nickname they last authed
   for. For example: I'm using the nick skold. I auth for skold2 and skold3.
   Then I /ns set noop off. NickServ assumes I want to set noop off for
   skold3, since I last id'd for it.
 */
NickInfo *get_nick (User *u)
{
    NickInfo *ni;

    if (!u)
    {
	log ("BUG: get_nick() called for null user");
	return NULL;
    }

    if (!u->lastnick)
	return NULL;

    ni = findnick (u->lastnick);

    if (ni)
	return ni;

    return NULL;
}

/* Check if the specified command is a password command. */
int ispass (char *text)
{
    char *cmd = strtok (text, " ");
    char *option = strtok (NULL, " ");

    /* Heres the idea. For half-security settings, we want
       to deny commands that deal with passwords. So this is
       where we check if the command is a password command.

       We also check that the command isn't HELP, since we
       don't want to deny help files.
     */
    if (!stricmp (cmd, "AUTH") ||
	!stricmp (cmd, "REGISTER") ||
	!stricmp (cmd, "IDENTIFY") ||
	!stricmp (cmd, "ID") ||
	!stricmp (cmd, "RELEASE") ||
	!stricmp (cmd, "LINK") ||
	!stricmp (cmd, "DROP") ||
	(!stricmp (cmd, "SET") && !stricmp (option, "PASS")))
	return 1;

    return 0;
}

/* Check if the command is a CTCP query. Return 1 if true, 0 if false. */
int isctcp (char *command)
{
    /* We use stristr to check for PINGs, since we don't know exactly what
       they'll look like.
     */
    if (stristr (command, "\1PING") || !stricmp (command, "\1VERSION\1"))
	return 1;

    return 0;
}

/* Is the given email valid? */
int validemail (char *email, const char *source)
{
    int i, valid = 1, chars = 0;

    /* Make sure it has @ and . */
    if (!strchr (email, '@') || !strchr (email, '.'))
	valid = 0;

    /* Check for other bad things */
    if (strchr (email, '$') || strchr (email, '/') ||
	strchr (email, ';') || strchr (email, '<') ||
	strchr (email, '>') || strchr (email, '&'))
    {
	globops (s_RootServ, ERR_HACK_EMAIL, source, email);
	valid = 0;
    }

    /* Make sure there are at least 6 characters besides the above
       mentioned @ and .
     */
    for (i = strlen (email) - 1; i > 0; i--)
	if (!(email[i] == '@' || email[i] == '.'))
	    chars++;

    if (chars < 6)
	valid = 0;

    return valid;
}

/* Generate a random number, for use as a key. */
unsigned long makekey (void)
{
    unsigned long i, j, k;

    /* This needs to be an unguessable number. Since most of these numbers
       will always be different at any given time, this should be completely
       unguessable. This could probably be a lot easier.. :)
     */
    i = rand() % (time (NULL) / usercnt + 1);
    j = rand() % (me.since * chancnt + 1);

    if (i > j)
	k = (i - j) + strlen (inbuf);
    else
	k = (j - i) + strlen (inbuf);

    /* Shorten or pad it to 9 digits. */
    if (k > 1000000000)
	k = k - 1000000000;
    if (k < 100000000)
	k = k + 100000000;

    return k;
}

/* Send the specified type of E-Mail.

   What is what we're sending to, a nickname or a channel name.

   param is an extra parameter; E-Mail, memo number, etc.

   If type ==:
     1 - Nick REGISTER E-Mail
     2 - Nick SENDPASS E-Mail
     3 - Channel REGISTER E-Mail
     4 - Channel SENDPASS E-Mail
     5 - MemoMail E-Mail
     6 - Nick E-Mail change confirmation
 */
void sendemail (const char *what, const char *param, int type)
{
    NickInfo *ni, *hni;
    ChanInfo *ci = NULL;
    MemoInfo *mi;
    Memo *m = NULL;
    char *email, *date, *subject, *tmp, *expire, *tmpex, *pass = NULL;
    char cmsg[256], nmsg[256], mmsg[256];
    char dBuf[2048], ret[2048], cmdbuf[512], timebuf[256], to[128], from[128];
    FILE *in, *out;
    time_t t;
    struct tm tm;
    int i, number = 0;

    /* Check if the file exists. Open it if it does. */
    if (type == 1)
	in = fopen (NSREGISTER, "r");
    else if (type == 2)
	in = fopen (NSSENDPASS, "r");
    else if (type == 3)
	in = fopen (CSREGISTER, "r");
    else if (type == 4)
	in = fopen (CSSENDPASS, "r");
    else if (type == 5)
	in = fopen (MSMEMOMAIL, "r");
    else
	in = fopen (NSCHANGE, "r");

    if (!in)
    {
	globops ("Warning! Can't find E-Mail file for %s. Aborting.",
	type == 1 ? "Nick REGISTER" : type == 2 ? "Nick SENDPASS" :
	type == 3 ? "Channel REGISTER" : type == 4 ? "Channel SENDPASS" :
	type == 5 ? "MemoMail" : type == 6 ? "E-Mail Change" : "Unknown");
	return;
    }

    if (type == 5)
	number = atoi (param);

    /* Get the nick we're talking to. */
    if (type == 1 || type == 2 || type == 5 || type == 6)
    {
	ni = findnick (what);

	if (!ni)
	    return;

	if (ni->host)
	    hni = findnick (ni->host);
	else
	    hni = ni;

	/* If this is a registration email, we want the key, not the
	   password.
	 */
	if (nsregistertype == 1 || nsregistertype == 3 || nsregistertype == 6)
	{
	    if (type == 2)
		pass = hni->pass;
	    else
		pass = itoa (hni->key);
	}

	expire = duration (nick_expire, 2);
    }
    else
    {
	ci = cs_findchan (what);
	ni = findnick (ci->founder);

	if (!ci || !ni)
	    return;

	if (ni->host)
	    hni = findnick (ni->host);
	else
	    hni = ni;

	if (type == 3)
	    pass = itoa (ci->key);
	else
	    pass = ci->pass;

	expire = duration (chan_expire, 2);
    }

    if (type == 5)
    {
	/* Find their memolist */
	mi = &hni->memos;

	/* Now get the memo we're supposed to be E-Mailing */
	for (i = 0, m = mi->memos; i < mi->memocnt; i++, m++)
	    if (m->number == number)
		break;

	if (!m)
	{
	    log ("sendemail(): Couldn't find specified memo %s for %s!",
		param, what);
	    return;
	}

	if (m->number != number)
	{
	    log ("sendemail(): Got wrong memo for %s!", what);
	    return;
	}
    }

    if ((!type == 6 && !hni->email) || (type == 6 && !hni->temp))
    {
	log ("sendemail(): Couldn't find E-Mail for %s! Bailing out.", what);
	return;
    }

    if (type == 6)
	email = hni->temp;
    else
	email = hni->email;

    tmpex = duration (tempexpire, 2);

    /* Set up the E-Mail headers */
    time (&t);
    tm = *gmtime (&t);
    strftime (timebuf, sizeof (timebuf) - 1, "%a, %d %b %Y %H:%M:%S %z", &tm);

    date = timebuf;
 
    if (type == 1 || type == 2 || type == 6)
	sprintf (from, "%s <%s@%s>", s_NickServ, ns.user, ns.host);
    else if (type == 3 || type == 4)
	sprintf (from, "%s <%s@%s>", s_ChanServ, cs.user, cs.host);
    else
	sprintf (from, "%s <%s@%s>", s_MemoServ, ms.user, ms.host);

    sprintf (to, "%s <%s>", hni->nick, email);
 
    if (type == 1)
	subject = "Nickname Registration";
    else if (type == 2 || type == 4)
	subject = "Password Retrieval";
    else if (type == 3)
	subject = "Channel Registration";
    else if (type == 5)
	subject = "MemoMail";
    else
	subject = "Change E-Mail Confirmation";

    /* Now set up the E-Mail */
    sprintf (cmdbuf, "%s %s", sendmail_path, email);
    out = popen (cmdbuf, "w");

    /* Set up the preferred form of MSGing. */
    snprintf (mmsg, sizeof (mmsg), "%s%s%s%s", haveserv_on == TRUE ? "" :
	"MSG ", s_MemoServ, haveserv_on == TRUE ? "" : securitysetting == 1 ?
	"@" : "", haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name :
	"");

    snprintf (nmsg, sizeof (nmsg), "%s%s%s%s", haveserv_on == TRUE ? "" :
	"MSG ", s_NickServ, haveserv_on == TRUE ? "" : securitysetting == 1 ?
	"@" : "", haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name :
	"");

    snprintf (cmsg, sizeof (cmsg), "%s%s%s%s", haveserv_on == TRUE ? "" :
	"MSG ", s_ChanServ, haveserv_on == TRUE ? "" : securitysetting == 1 ?
	"@" : "", haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name :
	"");

    fprintf (out, "From: %s\n", from);
    fprintf (out, "To: %s\n", to);
    fprintf (out, "Subject: %s\n", subject);
    fprintf (out, "Date: %s\n\n", date);

    while (fgets (dBuf, 2047, in))
    {
	if (strstr (dBuf, "$NW") || strstr (dBuf, "$CM") ||
	    strstr (dBuf, "$NM") || strstr (dBuf, "$MM") ||
	    strstr (dBuf, "$P") || strstr (dBuf, "$E") ||
	    strstr (dBuf, "$NS") || strstr (dBuf, "$CS") ||
	    strstr (dBuf, "$MS") || strstr (dBuf, "$TO") ||
	    strstr (dBuf, "$EX") || strstr (dBuf, "$N") ||
	    strstr (dBuf, "$C") || strstr (dBuf, "$MD") ||
	    strstr (dBuf, "$S") || strstr (dBuf, "$T"))
	{
	    /* Make a copy to play with */
	    strscpy (ret, dBuf, sizeof (ret));

	    /* Now replace each token with the right information. */
	    tmp = replace (ret, "$NW", network_name);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$MM", mmsg);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$NM", nmsg);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$CM", cmsg);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$P", pass);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$EX", expire);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$E", email);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$NS", s_NickServ);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$CS", s_ChanServ);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$MS", s_MemoServ);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$TO", tmpex);
	    strscpy (ret, tmp, sizeof (ret));
	    tmp = replace (ret, "$N", hni->nick);
	    strscpy (ret, tmp, sizeof (ret));

	    if (ci)
	    {
		tmp = replace (ret, "$C", ci->name);
		strscpy (ret, tmp, sizeof (ret));
	    }

	    if (m && type == 5)
	    {
		tmp = replace (ret, "$MD", get_time (m->time, 1));
		strscpy (ret, tmp, sizeof (ret));
		tmp = replace (ret, "$S", m->sender);
		strscpy (ret, tmp, sizeof (ret));
		tmp = replace (ret, "$T", m->text);
		strscpy (ret, tmp, sizeof (ret));
	    }

	    /* Now write out the line. */
	    fprintf (out, "%s", ret);

	    /* Now break out so we don't write the original line */
	    continue;
	}

	fprintf (out, "%s", dBuf);
    }

    fclose (in);
    pclose (out);
}

/* Check if the given nick is fully registered. If it isn't, we'll send a
   notice to source saying that is isn't, then return 0, indicating to the
   function calling this that it isn't fully registered, which will cause
   it to simply return. If the nick is fully regustered, we say nothing and
   return 1.
 */
int check_ns_auth (const char *whoami, const char *source, NickInfo *ni)
{
    if (ni->flags & NF_WAITAUTH)
    {
	notice (whoami, source, NS_NOT_AUTHED, ni->nick);
        return 0;
    }

    return 1;
}

/* Check if the given channel is fully registered. If it isn't, we'll send
   a notice to source saying that it isn't, then return 0, indicating to the
   function calling this that it isn't fully registered, which will cause
   it to simply return. If the channel is fully registered, we say nothing and
   return 1.
 */
int check_cs_auth (const char *whoami, const char *source, ChanInfo *ci)
{
    if (ci->flags & (CF_WAITING | CF_WAITAUTH))
    {
	notice (whoami, source, CS_NOT_AUTHED, ci->name);
	return 0;
    }

    return 1;
}

/* Put all psuedoclients in the JoinChan. */
void joinchan_join (void)
{
    if (!strlen (joinchan))
	return;

    if (rootserv_on == TRUE)
    {
#if defined (BAHAMUT) || defined (PROMETHEUS)
	send_cmd (s_RootServ, "SJOIN %ld %ld %s + :%s", time (NULL), time (NULL),
	    joinchan, s_RootServ);
#else
	send_cmd (s_RootServ, "JOIN %s", joinchan);
#endif
    }

    if (nickserv_on == TRUE)
    {
#if defined (BAHAMUT) || defined (PROMETHEUS)
	send_cmd (s_NickServ, "SJOIN %ld %ld %s + :%s", time (NULL), time (NULL),
	    joinchan, s_NickServ);
#else
	send_cmd (s_NickServ, "JOIN %s", joinchan);
#endif
    }

    if (chanserv_on == TRUE)
    {
#if defined (BAHAMUT) || defined (PROMETHEUS)
	send_cmd (s_ChanServ, "SJOIN %ld %ld %s + :%s", time (NULL), time (NULL),
	    joinchan, s_ChanServ);
#else
	send_cmd (s_ChanServ, "JOIN %s", joinchan);
#endif
    }

    if (memoserv_on == TRUE)
    {
#if defined (BAHAMUT) || defined (PROMETHEUS)
	send_cmd (s_MemoServ, "SJOIN %ld %ld %s + :%s", time (NULL), time (NULL),
	    joinchan, s_MemoServ);
#else
	send_cmd (s_MemoServ, "JOIN %s", joinchan);
#endif
    }

    if (globalnoticer_on == TRUE)
    {
#if defined (BAHAMUT) || defined (PROMETHEUS)
	send_cmd (s_GlobalNoticer, "SJOIN %ld %ld %s + :%s", time (NULL), time (NULL),
	    joinchan, s_GlobalNoticer);
#else
	send_cmd (s_GlobalNoticer, "JOIN %s", joinchan);
#endif
    }

    send_cmd (s_RootServ, "MODE %s +o%s%s%s%s %s %s %s %s %s", joinchan,
	nickserv_on == TRUE ? "o" : "",
	chanserv_on == TRUE ? "o" : "",
	memoserv_on == TRUE ? "o" : "",
	globalnoticer_on == TRUE ? "o" : "", s_RootServ,
	nickserv_on == TRUE ? s_NickServ : "",
	chanserv_on == TRUE ? s_ChanServ : "",
	memoserv_on == TRUE ? s_MemoServ : "",
	globalnoticer_on == TRUE ? s_GlobalNoticer : "");
}

/* Take all clients out of the JoinChan. */
void joinchan_part (void)
{
    if (rootserv_on == TRUE)
	send_cmd (s_RootServ, "PART %s", joinchan);

    if (nickserv_on == TRUE)
	send_cmd (s_NickServ, "PART %s", joinchan);

    if (chanserv_on == TRUE)
	send_cmd (s_ChanServ, "PART %s", joinchan);

    if (memoserv_on == TRUE)
	send_cmd (s_MemoServ, "PART %s", joinchan);

    if (globalnoticer_on == TRUE)
	send_cmd (s_GlobalNoticer, "PART %s", joinchan);
}

/* Notify all CSOps about a new object awaiting verification. */
void verify_notice (Verify *vf)
{
    User *u;
    int i;

    /* Go through the user list, notifying csops as we find them. */
    for (i = 0; i < 1024; i++)
	for (u = userlist[i]; u; u = u->next)
	    if (is_csop (u->nick))
	    {
		notice (s_ChanServ, u->nick, CS_WAITING_VERIFY, vf->name);
		notice (s_ChanServ, u->nick, CS_VERIFY_TOLIST, haveserv_on == TRUE ? "" :
		    "MSG ", s_ChanServ, haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    "@" : "", haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "",
		    "AUTH LIST");
	    }
}

/* Send a notice to all CSOps */
void noticecsops (const char *source, const char *fmt,...)
{
    User *u;
    int i;
    va_list args;
    char buf[BUFSIZE];

    va_start (args, fmt);

    for (i = 0; i < 1024; i++)
	for (u = userlist[i]; u; u = u->next)
	    if (is_csop (u->nick))
	    {
		snprintf (buf, sizeof (buf), "NOTICE %s :%s", u->nick, fmt);
		vsend_cmd (source, buf, args);
	    }

    va_end (args);
}

/* Check for any waiting verifications, and notify the csop if needed. */
void check_verifies (User *u)
{
    Verify *vf;
    ChanInfo *ci;
    int i, j = 0;

    for (i = 0; i < 1024; i++)
	for (vf = verifylist[i]; vf; vf = vf->next)
	{
	    ci = cs_findchan (vf->name);

	    if (ci)
		j++;
	    else
	    {
		/* Whups. It's not registered anymore. Take it out. */
		if (vf->prev)
		    vf->prev->next = vf->next;
		else
		    verifylist[VHASH(vf->name)] = vf->next;

		if (vf->next)
		    vf->next->prev = vf->prev;

		free (vf); 
	    }
	}

    if (j)
    {
	notice (s_ChanServ, u->nick, CS_VERIFY_WAITING, j);
	notice (s_ChanServ, u->nick, CS_VERIFY_TOLIST, haveserv_on == TRUE ? "" : "MSG ",
	    s_ChanServ, haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	    haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "", "AUTH LIST");
    }
}

/* Convert a text mode string to flags */
int32 string_to_flags (char *modes)
{
    int32 flags = 0;

    while (*modes)
    {
	int i;

	for (i = 0; cmodes[i].mode != 0 && cmodes[i].mode != *modes; i++);

	if (!cmodes[i].mode)
	    return 0;

	flags |= cmodes[i].flag;
	*modes++ = 0;
    }

    return flags;
}

/* Convert mode flags to a text mode string */
char *flags_to_string (int32 flags)
{
    static char buf[32];
    char *s = buf;
    int i;

    for (i = 0; cmodes[i].mode != 0; i++)
	if (flags & cmodes[i].flag)
	    *s++ = cmodes[i].mode;

    *s = 0;

    return buf;
}

/* Convert a mode character to a flag. */
int32 mode_to_flag (char c)
{
    int i;

    for (i = 0; cmodes[i].mode != 0 && cmodes[i].mode != c; i++);

    return cmodes[i].flag;
}

/* Convert a mode flag to a character. */
char mode_to_char (int32 f)
{
    int i;

    for (i = 0; cmodes[i].flag != 0 && cmodes[i].flag != f; i++);

    return cmodes[i].mode;
}

int32 prefix_to_flag (char c)
{
    int i;

    for (i = 0; cumodes[i].prefix != 0 && cumodes[i].prefix != c; i++);

    return cumodes[i].flag;
}

/* Send 'ERR_MORE_INFO' in the needed format. */
void errmoreinfo (const char *whoami, const char *source, char *text)
{
    notice (whoami, source, ERR_MORE_INFO, haveserv_on == TRUE ? "" : "MSG ",
	whoami, haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "", text);
}

/* Return the current time in milliseconds. */
uint32 time_msec (void)
{
#ifdef HAVE_GETTIMEOFDAY
    struct timeval tv;
    gettimeofday (&tv, NULL);
    return tv.tv_sec*1000 + tv.tv_usec/1000;
#else
    return time (NULL) * 1000;
#endif
}

#ifdef HAVE_GETTIMEOFDAY
/* Convert tv to ms */
int tv2ms (struct timeval *tv)
{   
    return (tv->tv_sec * 1000) + (int) (tv->tv_usec / 1000);
}
#endif

#if defined (BAHAMUT) || defined (PROMETHEUS) || defined (UNREAL3)
/* Go through every channel and nickname on the network.
   For nicknames, tell anyone who needs to, to identify.
   Also start collision timers.
   For each channel, do any modechanges and topic changes
   nessecary. Also kick any users as needed.
 */
void check_eb (void)
{
    ChanInfo *ci;
    Channel *c;
    User *u;
    struct c_userlist *cu;
    int ulev = 0;

    /* Check every user and call validate_user(). */
    for (u = firstuser (); u; u = nextuser ())
	validate_user (u);

    /* Go through every channel on the network. */
    for (c = firstchan (); c; c = nextchan ())
    {
	ci = cs_findchan (c->name);

	/* If the channel isn't registered, we have nothing
	   to do with it. Continue.
	 */
	if (!ci)
	    continue;

	/* Go through all the users on the channel, opping,
	   deopping, etc, as needed.
	 */
	for (cu = c->users; cu; cu = cu->next)
	{
	    u = cu->user;
	    ulev = get_access (u, ci);

	    /* FIRST: Do cs_join. This checks if the user should be
	       kicked for any reason, and also sends URLs and GREET
	       messages. If this returns 1, they were kicked for some
	       reason. Else they're fine, continue on.
	     */
	    if (cs_join (u, ci->name, 1))
		continue;

	    /* VOP? */
	    if (ulev == 1 && !(cu->mode & CMODE_v))
	    {
		send_cmode (s_ChanServ, c->name, "+v", u->nick);
		cu->mode |= CMODE_v;
	    }
#ifdef UNREAL3
	    /* HOP? */
	    if (ulev == 2 && !(cu->mode & CMODE_h))
	    {
		send_cmode (s_ChanServ, c->name, "+h", u->nick);
		cu->mode |= CMODE_h;
	    }
#endif
	    /* AOP or higher? */
	    if (ulev >= 3 && !(cu->mode & CMODE_o))
	    {
		send_cmode (s_ChanServ, c->name, "+o", u->nick);
		cu->mode |= CMODE_o;
	    }

	    /* Update the last used timestamp */
	    if (ulev)
		ci->lastused = time (NULL);

	    /* If they don't have any access, but VOPALL is on, voice them. */
	    if (!ulev && (ci->flags & CF_VOPALL))
	    {
		send_cmode (s_ChanServ, c->name, "+v", u->nick);
		cu->mode |= CMODE_v;
	    }

	    /* Now check if anyone has a mode they SHOULDN'T have. Only
	       do this if CF_SECURE is on, otherwise we don't bitch about
	       unauthed ops/etc.
	     */
	    if (ci->flags & CF_SECURE)
	    {
		if (ulev < 3 && (cu->mode & CMODE_o))
		{
		    send_cmode (s_ChanServ, c->name, "-o", u->nick);
		    cu->mode &= ~CMODE_o;
		}
#ifdef UNREAL3
		if (ulev < 2 && (cu->mode & CMODE_h))
		{
		    send_cmode (s_ChanServ, c->name, "-h", u->nick);
		    cu->mode &= ~CMODE_h;
		}
#endif
		if (ulev < 1 && (cu->mode & CMODE_v))
		{
		    send_cmode (s_ChanServ, c->name, "-v", u->nick);
		    cu->mode &= ~CMODE_v;
		}
	    }
	}

	/* check_modes() will take care of any
	   needed modechanges.
	 */
	check_modes (c->name);

	/* Check the current channel topic against the stored one.
	   If it differs, check the access of the person who set it.
	   If topiclock is off, or their access is sufficient, store
	   the new topic, else revert it.
	 */
	if (stricmp (ci->topic, c->topic) && !(ci->flags & CF_FROZEN))
	{
	    User *u = finduser (c->topicsetter);

	    ulev = 0;

	    if (u)
		ulev = get_access (u, ci);

	    if (!ulev || ulev < ci->topiclock)
	    {
		/* Revert the topic. */
		strscpy (c->topicsetter, ci->topicsetter, NICKLEN);
		strscpy (c->topic, ci->topic, TOPICLEN);
		c->topictime = ci->topictime;

		send_cmd (s_ChanServ, "TOPIC %s %s %lu :%s", c->name,
		    c->topicsetter, c->topictime, c->topic);
	    }
	    else
	    {
		/* No topiclock, sufficient access, etc. Store the new topic. */
		strscpy (ci->topicsetter, c->topicsetter, NICKLEN);
		strscpy (ci->topic, c->topic, TOPICLEN);
		ci->topictime = c->topictime;
	    }
	}
    }
}
#endif

/* Tell everyone identified to the given nickname that they
   have a new memo.
 */
void notify_new_memo (NickInfo *ni, char *from)
{
    User *u;

    for (u = firstuser (); u; u = nextuser ())
    {
	if (u->lastnick && !stricmp (ni->nick, u->lastnick))
	{
	    notice (s_MemoServ, u->nick, MS_NEW_MEMO, from);
	    notice (s_MemoServ, u->nick, MS_TO_LIST,
		haveserv_on == TRUE ? "" : "MSG ", s_MemoServ,
		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ?
		me.name : "");
	}
    }
}
