/* NickServ routines

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/services.h"

NickInfo *nicklist[1024];
int sstamp = 2147473598;

static void do_help (const char *source);
static void do_register (const char *source);
static void do_drop (const char *source);
static void do_info (const char *source);
static void do_identify (const char *source);
static void do_set (const char *source);
static void do_access (const char *source);
static void do_recover (const char *source);
static void do_link (const char *source);
static void do_unlink (const char *source);
static void do_links (const char *source);
static void do_chans (const char *source);
static void do_status (const char *source);
static void do_getpass (const char *source);
static void do_setpass (const char *source);
static void do_sendpass (const char *source);
static void do_freeze (const char *source);
static void do_unfreeze (const char *source);
static void do_hold (const char *source);
static void do_unhold (const char *source);
static void do_mark (const char *source);
static void do_unmark (const char *source);

static int delnick (NickInfo *ni);
static int is_on_access (User *u, NickInfo *ni);

/* Main NickServ routine. */
void nickserv (const char *source, char *buf)  
{
    char *cmd, *s;

    cmd = strtok (buf, " ");

    if (!cmd)
	return;
    
    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_NickServ, source, "PING %s", s);
	return;
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
    {
	ctcpreply (s_NickServ, source, "VERSION Cygnus IRC Services v%s",
	    version);
	return;
    }

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP",		H_NONE,		do_help},
	    {"REGISTER",	H_NONE,		do_register},
	    {"DROP",		H_NONE,		do_drop},
	    {"INFO",		H_NONE,		do_info},
	    {"*IDENTIFY",	H_NONE,		do_identify},
	    {"ID",		H_NONE,		do_identify},
	    {"SET",		H_NONE,		do_set},
	    {"ACCESS",		H_NONE,		do_access},
	    {"RECOVER",		H_NONE,		do_recover},
	    {"LINK",		H_NONE,		do_link},
	    {"UNLINK",		H_NONE,		do_unlink},
	    {"LINKS",		H_NONE,		do_links},
	    {"CHANS",		H_NONE,		do_chans},
	    {"STATUS",		H_NONE,		do_status},
	    {"GETPASS",		H_CSOP,		do_getpass},
	    {"SETPASS",		H_CSOP,		do_setpass},
	    {"SENDPASS",	H_CSOP,		do_sendpass},
	    {"FREEZE",		H_CSOP,		do_freeze},
	    {"UNFREEZE",	H_CSOP,		do_unfreeze},
	    {"HOLD",		H_SRA,		do_hold},
	    {"UNHOLD",		H_SRA,		do_unhold},
	    {"MARK",		H_SRA,		do_mark},
	    {"UNMARK",		H_SRA,		do_unmark},
	    {NULL}
	};
            
	if ((command = get_hash (s_NickServ, source, strupper (cmd),
				 hash_table)))
	{
	    /* Check for disabled or otherwise inapplicable commands. */
	    if (((command->process == do_link || command->process == do_links ||
		   command->process == do_unlink) && !maxlinks) ||
		(command->process == do_sendpass && sendmail_on == FALSE) ||
		(command->process == do_chans && chanserv_on == FALSE))
		notice (s_NickServ, source, ERR_UNKNOWN_COMMAND, haveserv_on == TRUE ? "" : 
		"MSG ", s_ChanServ, haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" :
		"", haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "",
		strupper (cmd));
	    else
		(*command->process) (source);
	}
    }
}

/* Return a help message. */
static void do_help (const char *source) 
{
    char *cmd = strtok (NULL, "");
            
    if (!cmd)
	nickserv_help_index (source);
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"REGISTER",	H_NONE,		nickserv_help_register},
	    {"DROP",		H_NONE,		nickserv_help_drop},
	    {"INFO",		H_NONE,		nickserv_help_info},
	    {"IDENTIFY",	H_NONE,		nickserv_help_identify},
	    {"ACCESS",		H_NONE,		nickserv_help_access},
	    {"LINK",		H_NONE,		nickserv_help_link},
	    {"UNLINK",		H_NONE,		nickserv_help_unlink},
	    {"LINKS",		H_NONE,		nickserv_help_links},
	    {"CHANS",		H_NONE,		nickserv_help_chans},
	    {"STATUS",		H_NONE,		nickserv_help_status},
	    {"RECOVER",		H_NONE,		nickserv_help_recover},
	    {"SET",		H_NONE,		nickserv_help_set},
	    {"GETPASS",		H_CSOP,		nickserv_help_getpass},
	    {"SETPASS",		H_CSOP,		nickserv_help_setpass},
	    {"SENDPASS",	H_CSOP,		nickserv_help_sendpass},
	    {"FREEZE",		H_CSOP,		nickserv_help_freeze},
	    {"UNFREEZE",	H_CSOP,		nickserv_help_unfreeze},
	    {"HOLD",		H_SRA,		nickserv_help_hold},
	    {"UNHOLD",		H_SRA,		nickserv_help_unhold},
	    {"MARK",		H_SRA,		nickserv_help_mark},
	    {"UNMARK",		H_SRA,		nickserv_help_unmark},
	    {NULL}
	};

	if ((command = get_help_hash (s_NickServ, source, strupper (cmd),
				      hash_table)))
	    (*command->process) (source);
    }
}

/* Check if the given user has access to the nickname. If not, 
   tell them to identify and start collision timers.
 */
int validate_user (User *u)
{
    NickInfo *ni, *hni;
    int access;

    if (debuglevel > 1)
	debug ("validate_user(): %s", u->nick);

    if (!(ni = findnick (u->nick)))
    {
	/* This nick isn't registered. Check if they have +r or +a. */
	if ((u->flags & UMODE_r) || (u->flags & UMODE_a))
	{
	    /* They may be a CSOp, but they're not registered. */
	    if (u->flags & UMODE_a)
	    {
		send_cmd (s_NickServ, "SVSMODE %s -ra", u->nick);
		u->flags &= ~(UMODE_r | UMODE_a);
	    }
	    else
	    {
		send_cmd (s_NickServ, "SVSMODE %s -r", u->nick);
		u->flags &= ~UMODE_r;
	    }
	}

	return 0;
    }

    /* If the nick is awaiting auth, bail out. */
    if (ni->flags & NF_WAITAUTH)
	return 0;

    /* If it's a linked nick, point to its host instead. */
    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    /* See if this nick is frozen (or linked to a frozen nick). Get them off of it if so. */
    if (hni->flags & NF_FROZEN)
    {
	notice (s_NickServ, u->nick, NS_INFO_FROZEN);

	if (hni->freezereason)
	    notice (s_NickServ, u->nick, NS_FROZEN_REASON, hni->freezereason);

	if (collidetype == 6 || (strlen (u->nick) == 30))
	    notice (s_NickServ, u->nick, NS_FROZEN_KILLED);
	else
#ifdef IRCUP9
	    notice (s_NickServ, u->nick, NS_FROZEN_KILLED);
#else
	    notice (s_NickServ, u->nick, NS_FROZEN_CHANGED);
#endif

	collide (ni, 0);

	return 0;
    }

    /* Check if they already have access to this nick */
    if (u->lastnick && !stricmp (u->lastnick, u->nick))
	return 0;

    /* Compare the server timestamp and servicesstamp. If these both match, it's as good 
       as an identify.
     */
    if ((hni->timestamp == u->timestamp) && (hni->sstamp == u->sstamp))
    {
	hni->lastseen = time (NULL);
	hni->usermask = smalloc (strlen (u->user) + strlen (u->host) + 2);
	sprintf (hni->usermask, "%s@%s", u->user, u->host);

	if (u->lastnick)
	    free (u->lastnick);

	u->lastnick = sstrdup (hni->nick);

	if (hni->real)
	    free (hni->real);

	hni->real = sstrdup (u->real);

	/* Set the SRA flag if SecureSRA is off */
	if (securesra_on == FALSE)
	    if (can_sra (hni->nick))
		u->flags |= UF_SRA;

#ifndef IRCUP9
	/* SRA-specific stuff */
	if (is_csop (u->nick))
	    send_cmd (s_NickServ, "SVSMODE %s +ra", u->nick);
	else
	    send_cmd (s_NickServ, "SVSMODE %s +r", u->nick);
#endif

#ifdef UNREAL3
	if (is_sra (u->nick) && securesra_on == FALSE)
	    send_cmd (s_NickServ, "SWHOIS %s :is a Services Root Admin", u->nick);
#endif

	/* Check for any new memos */
	check_memos (u);

	/* Check for waiting verifies */
	if (is_csop (u->nick))
	    check_verifies (u);

	/* Tell them about new news */
	if (hni->flags & NF_NEWS)
	{
	    FILE *f;

	    /* Make sure there still IS news to read */
	    f = fopen (NEWSFILE, "r");
 
	    if (f)
	    {
		notice (s_MemoServ, u->nick, MS_NEW_NEWS);
		notice (s_MemoServ, u->nick, MS_TO_READ_NEWS,
		    haveserv_on == TRUE ? "" : "MSG ", s_MemoServ,
		    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		    haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    me.name : "");
		fclose (f);
	    }
	    else
	    {
		/* The news file is gone now .. Turn the flag off. */
		hni->flags &= ~NF_NEWS;
	    }
	}

	return 1;
    }

    access = is_on_access (u, hni);

    if (!(hni->flags & NF_SECURE) && access)
    {
	hni->status |= NS_RECOGNIZED;
	hni->lastseen = time (NULL);

	if (hni->usermask)
	    free (hni->usermask);

	hni->usermask = smalloc (strlen (u->user) + strlen (u->host) + 2);
	sprintf (hni->usermask, "%s@%s", u->user, u->host);

	if (hni->real)
	    free (hni->real);

	hni->real = sstrdup (u->real);
	return 1;
    }

    notice (s_NickServ, u->nick, NS_NICK_OWNED);

    if (haveserv_on == TRUE)
	notice (s_NickServ, u->nick, NS_IDENTIFY, "", s_NickServ, "", "");
    else
	notice (s_NickServ, u->nick, NS_IDENTIFY, "MSG ", s_NickServ,
	securitysetting ? "@" : "", securitysetting ? me.name : "");

    if (((hni->flags & NF_ENFORCE) && !access) || (hni->flags & NF_SECURE))
    {
	notice (s_NickServ, u->nick, NS_60_SECONDS);
	add_ns_timeout (ni, TO_COLLIDE, 60);
	add_ns_timeout (ni, TO_MESS20, 40);
	add_ns_timeout (ni, TO_MESS40, 20);
    }

    return 0;
}

/* Is the given user's address on the given nick's access list? Return 1
   if so, 0 if not.
 */
static int is_on_access (User *u, NickInfo *ni)
{
    int i;
    char *buf;

    if (ni->accesscnt == 0)
	return 0;

    i = strlen (u->user);
    buf = smalloc (i + strlen (u->host) + 2);
    sprintf (buf, "%s@%s", u->user, u->host);
    strlower (buf + i + 1);

    for (i = 0; i < ni->accesscnt; i++)
    {
	if (match_wild_nocase (ni->access[i], buf))
	{
	    free (buf);
	    return 1;
	}
    }

    free (buf);
    return 0;
}

/* Release a held nickname from custody. */
void release (NickInfo *ni, int from_timeout)
{
    User *u;

    if (!from_timeout)
	del_ns_timeout (ni, TO_RELEASE, 1);

    /* Make sure this user isn't online. If someone's on with
       this nick, it isn't us, and some IRCds will generate an
       notice when they get the QUIT.
     */
    u = finduser (ni->nick);

    if (!u)
	send_cmd (ni->nick, "QUIT");

    ni->status &= ~NS_ENFORCED;
}

/* Collide a nickname. If we're on IRCu, we kill them from the network,
   otherwise we simply change their nick to whatever the conf setting for
   collisions is. Or, depending on the conf setting, we may kill them
   regardless of the ircd.
 */
void collide (NickInfo *ni, int from_timeout)
{
    User *u;
#ifndef IRCUP9
    char guest[NICKLEN];
    int guestnick, success = 0, tries = 0;
#endif

    u = finduser (ni->nick);

    if (!from_timeout)
    {
	del_ns_timeout (ni, TO_COLLIDE, 1);
	del_ns_timeout (ni, TO_MESS40, 1);
	del_ns_timeout (ni, TO_MESS20, 1);
    }

#ifndef IRCUP9
    if (collidetype == 6)
    {
	kill_user (s_NickServ, u->nick, NS_KILLED_ENFORCEMENT);

	ni->status |= NS_ENFORCED;
	u->flags |= UF_ENFORCED;

	return;
    }

    while (success == 0)
    {
	/* Check if their nickname is at the max length. If it is, some collidetypes
	   will fail, so we'll just kill them.
	 */
	if (collidetype == 3 || collidetype == 4 || collidetype == 5)
	{
#ifdef IRCUP9
	    if (strlen (u->nick) == 9)
#else
	    if (strlen (u->nick) == 30)
#endif
	    {
#ifndef IRCUP9
		kill_user (s_NickServ, u->nick, NS_KILLED_ENFORCEMENT);
#endif
		success = 1;
	    }
	}

	/* If we fail to get a usable nick 5 times in a row, kill them. */
	if (tries >= 5)
	{
#ifndef IRCUP9
	    kill_user (s_NickServ, u->nick, NS_KILLED_ENFORCEMENT);
#endif
	    success = 1;
	}

	if (collidetype == 1 || collidetype == 2 || collidetype == 3)
	{
	    /* Generate a random guest nick */
	    guestnick = 1+(int) (99999.0*random()/(RAND_MAX+10000.0));

	    /* If random() gives us a 4 digit number, add 10000 to it, bringing it into
	       the 5 digit range.
	     */
	    if (guestnick < 10000)
		guestnick += 10000;

	    if (collidetype == 1)
		snprintf (guest, sizeof (guest), "Guest%d", guestnick);
	    if (collidetype == 2)
		snprintf (guest, sizeof (guest), "User%d", guestnick);
	    if (collidetype == 3)
		snprintf (guest, sizeof (guest), "%s%d", u->nick, guestnick);
	}

	if (collidetype == 4)
	    snprintf (guest, sizeof (guest), "%s_", u->nick);

	if (collidetype == 5)
	    snprintf (guest, sizeof (guest), "%s-", u->nick);

	if (!finduser (guest))
	    success = 1;

	tries++;
    }

    /* If we killed them above, there's no sense sending this to them. */
    if (tries < 5)
    {
	if (from_timeout)
	{
	    notice (s_NickServ, u->nick, NS_NICK_CHANGED);

	    if (strlen (collision_url))
	    {
		notice (s_NickServ, u->nick, NS_COLLISION_INFO);
		notice (s_NickServ, u->nick, collision_url);
	    }
	}

	send_cmd (s_NickServ, "SVSNICK %s %s :%ld", u->nick, guest, time (NULL));
    }
#endif

#ifdef IRCUP9
    kill_user (s_NickServ, u->nick, NS_KILLED_ENFORCEMENT);
#endif

    ni->status |= NS_ENFORCED;
    u->flags |= UF_ENFORCED;
}

/* Cancel validation flags for a nick (i.e. when the user with that nick
   signs off or changes nicks). Also cancels any impending collides.
 */
void cancel_user (User * u)
{
    NickInfo *ni = findnick (u->nick);

    if (u->flags & UF_ENFORCED)
    {
	intro_user (u->nick, "enforcer", ns.host, "NickName Enforcement", "+i");
	u->flags &= ~UF_ENFORCED;

	if (ni)
	{
	    ni->status |= NS_ENFORCED;
	    add_ns_timeout (ni, TO_RELEASE, 60);
	}
    }
}

/* Delete a registered nickname. */
static int delnick (NickInfo *ni)
{
    NickInfo *tni;
    MemoInfo *mi;
    ChanInfo *ci, *next;
    ChanAccess *ca;
    int i;

#ifndef IRCUP9
    /* No longer registered. */
    if (finduser (ni->nick))
	send_cmd (s_NickServ, "SVSMODE %s -r", ni->nick);
#endif

    /* Delete any pending timeouts for this nick */
    del_ns_timeout (ni, -1, 1);

    /* Delete linked nicks */
    if (ni->links)
    {
	for (i = 0; i < ni->linkcnt; i++)
	{
	    if (ni->links[i])
	    {
		/* We store a list of the nicks linked to this one,
		   so we don't have to waste cpu hunting them down.
		   now we'll go through that list, find the nicks
		   linked to this one, and delete them.
		 */
		tni = findnick (ni->links[i]);

		/* This should NEVER fail. */
		if (tni)
		    delnick (tni);

		free (ni->links[i]);
	    }
	}

	free (ni->links);
    }

    /* Take us out of the host nick's info */
    if (ni->host)
    {
	char **links;

	tni = findnick (ni->host);

	/* This shouldn't fail.. better safe than sorry though. */
	if (tni)
	{
	    for (links = tni->links, i = 0; i < tni->linkcnt; links++, i++)
		if (!stricmp (*links, ni->nick))
		    break;

	    /* Set -r if this nick is in use */
	    if (finduser (tni->nick))
		send_cmd (s_NickServ, "SVSMODE %s -r", tni->nick);

	    if (*links)
		free (*links);

	    --tni->linkcnt;

	    if (i < tni->linkcnt)
		bcopy (links + 1, links, (tni->linkcnt - i) * sizeof (char *));

	    if (tni->linkcnt)
		tni->links = srealloc (tni->links, tni->linkcnt * sizeof
		    (char *));
	    else
	    {
		if (tni->links)
		    free (tni->links);
		tni->links = NULL;
	    }
	}

	free (ni->host);
    }

    /* Free last usermask */
    if (ni->usermask)
	free (ni->usermask);

    /* Free realname */
    if (ni->real)
	free (ni->real);

    /* Free email */
    if (ni->email)
	free (ni->email);

    /* Free url */
    if (ni->url)
	free (ni->url);

    /* Free name */
    if (ni->name)
	free (ni->name);

    /* Free sex */
    if (ni->sex)
	free (ni->sex);

    /* Free location */
    if (ni->location)
	free (ni->location);

    /* Temporary E-mail */
    if (ni->temp)
	free (ni->temp);

    /* Free access list */
    if (ni->access)
    {
	for (i = 0; i < ni->accesscnt; i++)
	{
	    if (ni->access[i])
		free (ni->access[i]);
	}

	free (ni->access);
    }

    /* Free memos */
    mi = &ni->memos;

    if (mi->memocnt)
    {
	for (i = 0; i < mi->memocnt; i++)
	    free (mi->memos[i].text);

	free (mi->memos);
	mi->memos = NULL;
	mi->memocnt = 0;
    }

    /* Check for any channels this nick is the successor for. Send a memo to the
       founder.
     */
    if (ni->successor)
    {
	ChanInfo *ci;

	for (i = 0; i < ni->successorcnt; i++)
	{
	    if (ni->successor[i])
	    {
		/* We store a list of channels we're the successor for so that we can
		   notify the current founder of the channel when the successor's nickname
		   expires. May seem like a waste, but the alternative is searching every
		   registered channel to get this information, or not notifying them at all.
		 */
		ci = cs_findchan (ni->successor[i]);

		/* This shouldn't fail. */
		if (ci)
		{
		    NickInfo *fni = findnick (ci->founder);
		    MemoInfo *mi;
		    Memo *m;
		    char buf[256];
		    int j;

		    mi = &fni->memos;
		    mi->memocnt++;
		    mi->memos = srealloc (mi->memos, sizeof (Memo) * mi->memocnt);
		    m = &mi->memos[mi->memocnt - 1];

		    /* It's 'from' ChanServ */
		    strscpy (m->sender, s_ChanServ, NICKLEN);

		    if (mi->memocnt > 1)
		    {
			m->number = m[-1].number + 1;

			if (m->number < 1)
			    for (j = 0; j < mi->memocnt; j++)

			mi->memos[i].number = i + 1;
		    }
		    else
			m->number = 1;

		    snprintf (buf, sizeof (buf), CS_SUCCESSOR_GONE_MEMO, ci->name, ni->nick);
		    m->time = time (NULL);
		    m->text = sstrdup (buf);
		    m->flags = MF_UNREAD;

		    /* Now tell anyone identified to this nick that they have a new memo. */
		    notify_new_memo (fni, s_ChanServ);
		}

		free (ni->successor[i]);
	    }
	}

	free (ni->successor);
    }

    /* Check for successors to any owned channels. */
    if (ni->chans)
    {
	for (i = 0; i < ni->chancnt; i++)
	{
	    if (ni->chans[i])
	    {
		/* Like links, we store a list of channels this nickname
		   owns on the nick to save time. Since this nick is being
		   dropped, we'll delete those channels. If they have a
		   successor set, we'll check if that nick is registered and
		   under any channel limits, if everything checks out, we'll
		   transfer it to them.
		 */
		ci = cs_findchan (ni->chans[i]);

		/* This shouldn't fail. */
		if (ci)
		{
		    int delete = 1;

		    if (ci->successor)
		    {
			NickInfo *sni = findnick (ci->successor);
			MemoInfo *mi;
			Memo *m;
			char buf[256];
			int j;

			if (sni)
			{
			    /* Woo, the successor's nick is still registered, but do
			       they have enough channel spots left?
			     */
			    if (sni->chancnt < chanlimit)
			    {
				/* Kickass. Transfer it to them. */
				strscpy (ci->founder, sni->nick, NICKLEN);

				/* We don't want to delete this channel, now */
				delete = 0;

				/* No more successor */
				free (ci->successor);
				ci->successor = NULL;

				/* Add it to their channel list */
				++sni->chancnt;
				sni->chans = srealloc (sni->chans, sizeof (char *) *sni->chancnt);
				sni->chans[sni->chancnt - 1] = sstrdup (ci->name);

				/* And finally, tell the new owner about it. */
				mi = &sni->memos;
				mi->memocnt++;
				mi->memos = srealloc (mi->memos, sizeof (Memo) *
				    mi->memocnt);
				m = &mi->memos[mi->memocnt - 1];

				/* It's 'from' ChanServ */
				strscpy (m->sender, s_ChanServ, NICKLEN);

				if (mi->memocnt > 1)
				{
				    m->number = m[-1].number + 1;

				    if (m->number < 1)
					for (j = 0; j < mi->memocnt; j++)

				    mi->memos[j].number = j + 1;
				}
				else
				    m->number = 1;

				snprintf (buf, sizeof (buf), CS_SUCCESSOR_MEMO, ci->name);

				m->time = time (NULL);
				m->text = sstrdup (buf);
				m->flags = MF_UNREAD;

				/* Tell anyone id'd to this nick they have a new memo */
				notify_new_memo (sni, s_ChanServ);
 			    }
			}
		    }

		    if (delete)
			delchan (ci);
		}
	    }
	}
    }

    /* Go through each channel's access list, and remove this person
       from them as needed.
     */
    for (ci = firstci (); ci; ci = next)
    {
	next = nextci ();

	for (ca = ci->access, i = 0; i < ci->accesscnt; ca++, i++)
	{
	    if (!ca->level)
		continue;

	    if (!stricmp (ni->nick, ca->mask))
	    {
		if (ca->mask)
		    free (ca->mask);

		ca->level = 0;
	    }
	}
    }

    if (ni->prev)
	ni->prev->next = ni->next;
    else
	nicklist[NSHASH(ni->nick)] = ni->next;

    if (ni->next)
	ni->next->prev = ni->prev;

    free (ni);

    return 1;
}

/* Check all nicknames and expire any nessecary */
void expire_nicks ()
{
    User *u;
    NickInfo *ni, *next;
    int j = 0;

    if (logupdates_on == TRUE)
	log_sameline (1, "NS:SYNC: Checking expires... ");

    for (ni = firstni (); ni; ni = next)
    {
	next = nextni ();

	/* While we're here, check for any nicknames that are online.
	   if they are, update their lastseen time. This stops them from
	   being expired in error if they've been online longer than
	   nick_expire.
	 */
	if ((u = finduser (ni->nick)))
	    if (u && u->lastnick && !stricmp (u->lastnick, ni->nick))
		ni->lastseen = time (NULL);

	/* Only non-linked nicks and non-held nicks */
	if (!(ni->flags & NF_LINKED) && !(ni->flags & NF_HELD))
	{
	    if ((time (NULL) >= ni->lastseen + nick_expire) ||
		(ni->flags & NF_WAITAUTH && time (NULL) >=
		ni->registered + tempexpire))
	    {
		if (ni->linkcnt)
		    log ("NS:SYNC:EXPIRE: %s (%d)", ni->nick,
			ni->linkcnt);
		else
		    log ("NS:SYNC:EXPIRE: %s", ni->nick);

		delnick (ni);

		j++;
	    }
	}
    }

    /* Only log this if we have something to report, or if we should
       log it regardless.
     */
    if (logupdates_on == TRUE)
	log_sameline (0, "%d nicks expired.", j);
    else if (logupdates_on == FALSE && j > 0)
	log ("NS:SYNC:EXPIRE: %d nicks expired.", j);
}

/* Register a nickname. */
static void do_register (const char *source)
{
    NickInfo *ni, *tni, **list;
    User *u = finduser (source);
    FILE *f;
    char *param1 = strtok (NULL, " ");
    char *param2 = strtok (NULL, " ");
    char *param3 = strtok (NULL, " ");
    int i, found = 0, reason = 0;
    time_t now = time (NULL);

    if (!u)
	return;

    /* Check here if the're completing registration from E-Mail AUTH */
    if (param1 && param2 && (nsregistertype == 1 || nsregistertype == 3 ||
	nsregistertype == 6))
    {
	ni = findnick (param1);

	/* Is this nick waiting for completion? */
	if (ni && (ni->flags & NF_WAITAUTH))
	{
	    if (ni->key && ni->key == atol (param2))
	    {
		notice (s_NickServ, source, NS_REGISTERED);
		notice (s_NickServ, source, RPL_PASSWORD_SET, ni->pass);

		if (strlen (nsregister_url))
		{
		    notice (s_NickServ, source, NS_REGISTER_INFO);
		    notice (s_NickServ, source, nsregister_url);
		}

		/* Tell them about new news.. This is a new nick, we'll assume
		   they haven't read the news. If we find the newsfile, we'll
		   turn the flag on for them.
		 */
		f = fopen (NEWSFILE, "r");

		if (f)
		{
		    /* News found, tell them about it */
		    notice (s_MemoServ, source, MS_NEW_NEWS);
		    notice (s_MemoServ, source, MS_TO_READ_NEWS,
			haveserv_on == TRUE ? "" : "MSG ", s_MemoServ,
			haveserv_on == TRUE ? "" : securitysetting == 1 ?
			"@" : "", haveserv_on == TRUE ? "" :
			securitysetting == 1 ? me.name : "");
		    fclose (f);

		    /* Now turn the NF_NEWS flag on for them */
		    ni->flags |= NF_NEWS;
		}

		/* Turn the key off */
		ni->key = 0;

		/* No longer waiting */
		ni->flags &= ~NF_WAITAUTH;

		/* Identified */
		if (u->lastnick)
		    free (u->lastnick);

		u->lastnick = sstrdup (ni->nick);

		/* Set the SRA flag if SecureSRA is off */
		if (securesra_on == FALSE)
		{
		    if (can_sra (ni->nick))
			u->flags |= UF_SRA;

#ifndef IRCUP9
		    if (is_sra (source))
		    {
			send_cmd (s_NickServ, "SVSMODE %s +ra", source);
#ifdef UNREAL3
			send_cmd (s_NickServ,
			    "SWHOIS %s :is a Services Root Admin", source);
#endif
		    }
		    else
			send_cmd (s_NickServ, "SVSMODE %s +r", source);
#endif
		}

		return;
	    }
	}
    }

    /* Check here first if this nick is a Guest or User nick. We don't want
       people registering those.
     */
    if (match_wild_nocase ("Guest*", u->nick) ||
	match_wild_nocase ("User*", u->nick))
    {
	notice (s_NickServ, source, ERR_NOT_GUEST, strstr (u->nick, "Guest") ?
	    "Guest" : "User");
	return;
    }

    /* Check that we have all needed parameters and that they're valid. */
    if (!nsregistertype && !param1)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "REGISTER Password");
	errmoreinfo (s_NickServ, source, "REGISTER");
	return;
    }

    /* REGISTER E-Mail */
    if (nsregistertype == 1)
    {
	if (!param1)
	{
	    notice (s_NickServ, source, RPL_SYNTAX, "REGISTER E-Mail");
	    errmoreinfo (s_NickServ, source, "REGISTER");
	    return;
	}

	if (!validemail (param1, source))
	{
	    notice (s_NickServ, source, ERR_INVALID, "E-Mail", "");
	    return;
	}
    }

    /* REGISTER Pass E-Mail */
    if (nsregistertype == 2 || nsregistertype == 3)
    {
	if (!param1 || !param2)
	{
	    notice (s_NickServ, source, RPL_SYNTAX, "REGISTER Password E-Mail");
	    errmoreinfo (s_NickServ, source, "REGISTER");
	    return;
	}

	if (!validemail (param2, source))
	{
	    notice (s_NickServ, source, ERR_INVALID, "E-Mail", "");
	    return;
	}
    }

    /* REGISTER Pass Pass */
    if (nsregistertype == 4)
    {
	if (!param1 || !param2)
	{
	    notice (s_NickServ, source, RPL_SYNTAX, "REGISTER Password Password");
	    errmoreinfo (s_NickServ, source, "REGISTER");
	    return;
	}

	if (strcmp (param1, param2))
	{
	    notice (s_NickServ, source, ERR_PASS_DONT_MATCH);
	    return;
	}
    }

    /* REGISTER Pass Pass E-Mail */
    if (nsregistertype == 5 || nsregistertype == 6)
    {

	if (!param1 || !param2 || !param3)
	{
	    notice (s_NickServ, source, RPL_SYNTAX,
		"REGISTER Password Password E-Mail");
	    errmoreinfo (s_NickServ, source, "REGISTER");
	    return;
	}

	if (strcmp (param1, param2))
	{
	    notice (s_NickServ, source, ERR_PASS_DONT_MATCH);
	    return;
	}

	if (!validemail (param3, source))
	{
	    notice (s_NickServ, source, ERR_INVALID, "E-Mail", "");
	    return;
	}
    }

    /* No register floods! */
    if (nick_delay && (now - u->lastnreg < nick_delay))
    {
	notice (s_NickServ, source, NS_WAIT_REG,
	    duration (u->lastnreg - now + nick_delay, 2));

	/* Pimp nick linking to them */
	if (maxlinks)
	{
	    notice (s_NickServ, source, NS_TRY_LINK);
	    notice (s_NickServ, source, NS_HELP_LINK,
		haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	}

	return;
    }

    /* Is it a Services nick? */
    if (is_one_of_mine (source))
    {
	notice (s_NickServ, source, NS_NOT_SERVICES_NICK);
	return;
    }

    /* Is it already registered? */
    if ((ni = findnick (source)))
    {
	notice (s_NickServ, source, NS_ALREADY_REGISTERED);
	return;
    }

    /* Password sanity checks */
    if (nsregistertype != 1)
    {
	if (strlen (param1) < 5)
	    reason = 1;
	if (strlen (param1) > PASSWDLEN)
	    reason = 2;
	if (!stricmp (source, param1))
	    reason = 3;
	if (!stricmp ("password", param1))
	    reason = 4;

	if (reason)
	{
	    notice (s_NickServ, source, ERR_BETTER_PASS);

	    if (reason == 1)
		notice (s_NickServ, source, ERR_PASS_REJECTED_SHORT);
	    if (reason == 2)
		notice (s_NickServ, source, ERR_PASS_REJECTED_LONG);
	    if (reason == 3)
		notice (s_NickServ, source, ERR_PASS_REJECTED_NICK);
	    if (reason == 4)
		notice (s_NickServ, source, ERR_PASS_REJECTED_WEAK);

	    errmoreinfo (s_NickServ, source, "REGISTER");
	    return;
	}
    }

    /* Check for an email limit */
    if (emaillimit && !(!nsregistertype || nsregistertype == 4))
    {
	char *email = "";

	if (nsregistertype == 1)
	    email = param1;
	if (nsregistertype == 2 || nsregistertype == 3)
	    email = param2;
	if (nsregistertype == 5 || nsregistertype == 6)
	    email = param3;

	for (i = 0; i < 1024; i++)
	    for (tni = nicklist[i]; tni; tni = tni->next)
		if (tni->email && !stricmp (tni->email, email))
		    ++found;

	if (found && found >= emaillimit)
	{
	    notice (s_NickServ, source, NS_TOO_MANY_EMAILS, email);
	    return;
	}
    }

    /* Register the NickName. */
    ni = scalloc (sizeof (NickInfo), 1);
    strscpy (ni->nick, source, NICKLEN);

    if (nsregistertype != 1)
	strscpy (ni->pass, param1, PASSWDLEN);

    /* We'll make a pw */
    if (nsregistertype == 1 || nsregistertype == 3 || nsregistertype == 6)
	ni->key = makekey ();

    /* They don't have a password. We'll use the key as their pw. */
    if (nsregistertype == 1)
	strscpy (ni->pass, itoa (ni->key), PASSWDLEN);

    ni->registered = ni->lastseen = now;
    ni->memos.memolimit = memobox_size;
    ni->real = sstrdup (u->real);
    ni->usermask = smalloc (strlen (u->user) + strlen (u->host) + 2);
    sprintf (ni->usermask, "%s@%s", u->user, u->host);
    ni->accesscnt = 1;
    ni->access = smalloc (sizeof (char *));
    ni->access[0] = ni->usermask;

    if (defnickflags)
	ni->flags = defnickflags;

    if (nsregistertype == 1)
	ni->email = sstrdup (param1);

    if (nsregistertype == 2 || nsregistertype == 3)
	ni->email = sstrdup (param2);

    if (nsregistertype == 5 || nsregistertype == 6)
	ni->email = sstrdup (param3);

    if (u->lastnick)
	free (u->lastnick);

    u->lastnick = sstrdup (ni->nick);
    u->lastnreg = now;

    list = &nicklist[NSHASH(ni->nick)];
    ni->next = *list;
    if (*list)
	(*list)->prev = ni;
    *list = ni;

    /* Make note of the occurance. */
    log ("NS:REGISTER: %s!%s@%s", source, u->user, u->host);
    snoop (s_NickServ, "NS:REGISTER: %s!%s@%s", source, u->user,
	u->host);

    /* If this was an E-Mail auth register, we'll tell them to check their
       email, email the instructions, and bail. We don't need to go beyond
       this point.
     */
    if (nsregistertype == 1 || nsregistertype == 3 || nsregistertype == 6)
    {
	/* Is this email being abused? */
	if (check_email (source, ni->email))
	{
	    notice (s_NickServ, source, ERR_EMAIL_USED);
	    delnick (ni);
	    return;
	}

	sendemail (ni->nick, itoa (ni->key), 1);

	notice (s_NickServ, source, NS_REGISTERED_TEMP);
	notice (s_NickServ, source, RPL_EMAIL_INSTRUCTIONS);
	notice (s_NickServ, source, NS_TEMP_EXPIRE, duration (tempexpire, 2));

	/* Set a flag so we know we're waiting for them */
	ni->flags |= NF_WAITAUTH;

	return;
    }

    /* Set the SRA flag if SecureSRA is off */
    if (securesra_on == FALSE)
    {
	if (can_sra (ni->nick))
	    u->flags |= UF_SRA;

#ifndef IRCUP9
	if (is_sra (source))
	{
	    send_cmd (s_NickServ, "SVSMODE %s +ra", source);
#ifdef UNREAL3
	    send_cmd (s_NickServ,
		"SWHOIS %s :is a Services Root Admin", source);
#endif
	}
	else
	    send_cmd (s_NickServ, "SVSMODE %s +r", source);
#endif
    }

    /* Tell them we were successful. */
    notice (s_NickServ, source, NS_REGISTERED);
    notice (s_NickServ, source, RPL_PASSWORD_SET, ni->pass);

    if (strlen (nsregister_url))
    {
	notice (s_NickServ, source, NS_REGISTER_INFO);
	notice (s_NickServ, source, nsregister_url);
    }

    /* Tell them about new news.. This is a new nick, we'll assume they
       haven't read the news. If we find the newsfile, we'll turn the flag
       on for them.
     */
    f = fopen (NEWSFILE, "r");

    if (f)
    {
	/* News found, tell them about it */
	notice (s_MemoServ, source, MS_NEW_NEWS);
	notice (s_MemoServ, source, MS_TO_READ_NEWS,
	    haveserv_on == TRUE ? "" : "MSG ", s_MemoServ,
	    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	    haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	fclose (f);

	/* Now turn the NF_NEWS flag on for them */
	ni->flags |= NF_NEWS;
    }
}

/* Drop a registered nickname */
static void do_drop (const char *source)
{
    NickInfo *ni, *hni;
    User *u = finduser (source);
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");

    if (!u)
	return;

    if (!nick || !pass)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "DROP NickName Password");
	errmoreinfo (s_NickServ, source, "DROP");
	return;
    }

    ni = findnick (nick);

    if (!ni)
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (ni->flags & NF_FROZEN)
    {
	notice (s_NickServ, source, NS_NICK_FROZEN, ni->nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Linked nicks don't have a password. Point to host instead */
    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    if (!strcmp (pass, hni->pass))
    {
	if (ni->linkcnt)
	{
	    log ("NS:DROP: %s %s!%s@%s (%d)", ni->nick, source, u->user, u->host,
		ni->linkcnt);
	    snoop (s_NickServ, "NS:DROP: %s %s!%s@%s (%d)", ni->nick, source,
		u->user, u->host, ni->linkcnt);
	}
	else
	{
	    log ("NS:DROP: %s %s!%s@%s", ni->nick, source, u->user, u->host);
	    snoop (s_NickServ, "NS:DROP: %s %s!%s@%s", ni->nick, source, u->user,
		u->host);
	}

	delnick (ni);

	notice (s_NickServ, source, NS_DROPPED, ni->nick);
    }
    else
    {
	/* Passwords don't match, tell them and update their passfail count */
	++u->passfail;

	notice (s_NickServ, source, RPL_WRONG_PASS, ni->nick);

	/* If they've exceeded the limit, kill them. */
	if (u->passfail >= passfails)
	{
	    snoop (s_NickServ, "NS:DROP:BP: %s %s!%s@%s (Killed)",
		ni->nick, source, u->user, u->host);
	    log ("NS:DROP:BP: %s %s!%s@%s (Killed)", ni->nick,
		source, u->user, u->host);
	    kill_user (s_NickServ, source, RPL_KILLED_PASSFAIL);
	    return;
	}

	if (u->passfail >= (passfails - 1))
	    notice (s_NickServ, source, RPL_LAST_TRY);
	snoop (s_NickServ, "NS:DROP:BP: %s %s!%s@%s (%d/%d)", ni->nick,
	    source, u->user, u->host, u->passfail, passfails);
	log ("NS:DROP:BP: %s %s!%s@%s (%d/%d)",
	    ni->nick, source, u->user, u->host, u->passfail, passfails);

	return;
    }
}

/* Link a nickname to another nickname */
static void do_link (const char *source)
{
    NickInfo *ni, *lni, **list;
    User *u = finduser (source);
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    time_t now = time (NULL);

    if (!u)
	return;

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "LINK NickName [Password]");
	errmoreinfo (s_NickServ, source, "LINK");
	return;
    }

    /* Check here first if this nick is a Guest or User nick. We don't want
       people registering those.
     */
    if (match_wild_nocase ("Guest*", u->nick) ||
        match_wild_nocase ("User*", u->nick) ||
	match_wild_nocase ("Guest*", nick) ||
	match_wild_nocase ("User*", nick))
    {
	notice (s_NickServ, source, ERR_NOT_GUEST, strstr (u->nick, "Guest") ?
	    "Guest" : "User");
	return;
    }

    /* No register floods! */
    if (nick_delay && (now - u->lastnreg < nick_delay))
    {
	notice (s_NickServ, source, NS_WAIT_REG,
	    duration (u->lastnreg - now + nick_delay, 2));
	return;
    }

    lni = findnick (nick);
    ni = get_nick (u);

    if (ni)
	if (!check_ns_auth (s_NickServ, source, ni))
	    return;

    /* lni->nick and ni->nick might be the same. We don't want to link
       a nick to itself. Since they might be authed for the specified
       nick and using an unregistered nick, we'll set ni to source.
     */
    if ((ni && lni && !stricmp (lni->nick, ni->nick)))
    {
	ni = findnick (source);

	/* Is it STILL the same nick? */
	if ((ni && lni && !stricmp (lni->nick, ni->nick)))
	{
	    notice (s_NickServ, source, NS_LINK_SELF);
	    return;
	}
    }

    /* Neither nickname is registered. */ 
    if (!lni && !ni)
    {
	notice (s_NickServ, source, NS_IDENTIFY_ONE);
	return;
    }

    /* The specified nickname isn't registered, but the source one is. */
    if (!lni && ni)
    {
	/* Since they want to link an unregistered nickname to their
	   current nickname, we'll skip the password and just make
	   sure they're identified.
	 */
	if (!u->lastnick || stricmp (ni->nick, u->lastnick))
	{
	    notice (s_NickServ, source, ERR_NEED_IDENT, "LINK");
	    return;
	}

	/* Make sure we're not being asked to link to a linked nick */
	if (ni->flags & NF_LINKED)
	{
	    /* Beh. It's linked. Grab the host nick instead. */
	    ni = findnick (ni->host);
	}

	/* Do they have too many? */
	if (ni->linkcnt >= maxlinks)
	{
	    notice (s_NickServ, source, NS_MAX_LINKS, maxlinks);
	    return;
	}

	/* Make note of it on the host nick */
	++ni->linkcnt;
	u->lastnreg = now;
	ni->links = srealloc (ni->links, sizeof (char *) *ni->linkcnt);
	ni->links[ni->linkcnt - 1] = sstrdup (nick);

	/* Linked nicknames are normally registered nicks, except they have
	   the LINKED flag, and have the host nickname stored. Thus, we
	   have to register it here.
	 */
	lni = scalloc (sizeof (NickInfo), 1);
	strscpy (lni->nick, nick, NICKLEN);
	lni->host = sstrdup (ni->nick);
	lni->flags = NF_LINKED;

	list = &nicklist[NSHASH(lni->nick)];
	lni->next = *list;
	if (*list)
	    (*list)->prev = lni;
	*list = lni;

	notice (s_NickServ, source, NS_NICKS_LINKED, lni->nick, ni->nick);
	log ("NS:LINK: %s %s %s!%s@%s", lni->nick, ni->nick,
	    source, u->user, u->host);
	snoop (s_NickServ, "NS:LINK: %s %s %s!%s@%s", lni->nick,
	    ni->nick, source, u->user, u->host);
	return;
    }

    /* The source nickname isn't registered, but the specified one is. */
    if (lni && !ni)
    {
	if (!pass)
	{
	    notice (s_NickServ, source, RPL_SYNTAX, "LINK NickName Password");
	    errmoreinfo (s_NickServ, source, "LINK");
	    return;
	}

	/* Check the passwords */
	if (strcmp (pass, lni->pass))
	{
	    /* Passwords don't match, tell them and update their passfail
	       count
	     */
	    ++u->passfail;

	    notice (s_NickServ, source, RPL_WRONG_PASS, lni->nick);

	    /* If they've exceeded the limit, kill them. */
	    if (u->passfail >= passfails)
	    {
		snoop (s_NickServ,
		    "NS:LINK:BP: %s %s %s!%s@%s (Killed)",
		    nick, lni->nick, source, u->user, u->host);
		log ("NS:LINK:BP: %s %s %s!%s@%s (Killed)",
		    nick, lni->nick, source, u->user, u->host);
		kill_user (s_NickServ, source, RPL_KILLED_PASSFAIL);
		return;
	    }

	    if (u->passfail >= (passfails - 1))
		notice (s_NickServ, source, RPL_LAST_TRY);

	    snoop (s_NickServ,
		"NS:LINK:BP: %s %s %s!%s@%s (%d/%d)", nick,
		lni->nick, source, u->user, u->host, u->passfail, passfails);

	    log ("NS:LINK:BP: %s %s %s!%s@%s (%d/%d)", nick,
		lni->nick, source, u->user, u->host, u->passfail, passfails);

	    return;
	}

	/* Make sure we're not being asked to link to a linked nick */
	if (lni->flags & NF_LINKED)
	{
	    /* Beh. It's linked. Grab the host nick instead. */
	    lni = findnick (lni->host);
	}

	/* Do they have too many? */
	if (lni->linkcnt >= maxlinks)
	{
	    notice (s_NickServ, source, NS_MAX_LINKS_OTHER, lni->nick, maxlinks);
	    return;
	}

	if (!check_ns_auth (s_NickServ, source, lni))
	    return;

	++lni->linkcnt;
	u->lastnreg = now;
	lni->links = srealloc (lni->links, sizeof (char *) *lni->linkcnt);
	lni->links[lni->linkcnt - 1] = sstrdup (source);

	ni = scalloc (sizeof (NickInfo), 1);
	strscpy (ni->nick, source, NICKLEN);
	ni->host = sstrdup (lni->nick);
	ni->flags = NF_LINKED;

	list = &nicklist[NSHASH(ni->nick)];
	ni->next = *list;
	if (*list)
	    (*list)->prev = ni;
	*list = ni;

	/* They've essentially identified for this nickname, so we'll
	   consider them as such.
	 */
	if (u->lastnick)
	    free (u->lastnick);

	u->lastnick = sstrdup (ni->nick);

	notice (s_NickServ, source, NS_NICKS_LINKED, ni->nick, lni->nick);
	log ("NS:LINK: %s %s %s!%s@%s", ni->nick, lni->nick,
	    source, u->user, u->host);
	snoop (s_NickServ, "NS:LINK: %s %s %s!%s@%s", ni->nick,
	    lni->nick, source, u->user, u->host);
	return;
    }

    /* Both nicknames are registered. We'll merge them. If the user read
       the help files (which they should have) they'll know that the
       specified nickname is considered the host.
     */
    if (lni && ni)
    {
	char tnick[NICKLEN];

	if (!pass)
	{
	    notice (s_NickServ, source, RPL_SYNTAX, "LINK NickName Password");
	    errmoreinfo (s_NickServ, source, "LINK");
	    return;
	}

	/* Check the passwords */
	if (strcmp (pass, lni->pass))
	{
	    /* Passwords don't match, tell them and update their passfail
	       count
	     */
	    ++u->passfail;

	    notice (s_NickServ, source, RPL_WRONG_PASS, lni->nick);

	    /* If they've exceeded the limit, kill them. */
	    if (u->passfail >= passfails)
	    {
		snoop (s_NickServ,
		    "NS:LINK:BP: %s %s %s!%s@%s (Killed)", ni->nick,
		    lni->nick, source, u->user, u->host);
		log ("NS:LINK:BP: %s %s %s!%s@%s (Killed)",
		    ni->nick, lni->nick, source, u->user, u->host);
		kill_user (s_NickServ, source, RPL_KILLED_PASSFAIL);
		return;
	    }

	    if (u->passfail >= (passfails - 1))
		notice (s_NickServ, source, RPL_LAST_TRY);
	    snoop (s_NickServ,
		"NS:LINK:BP: %s %s %s!%s@%s (%d/%d)", ni->nick,
		lni->nick, source, u->user, u->host, u->passfail, passfails);
	    log ("NS:LINK:BP: %s %s %s!%s@%s", ni->nick,
		lni->nick, source, u->user, u->host, u->passfail, passfails);

	    return;
	}

	/* Make sure we're not being asked to link to a linked nick */
	if (lni->flags & NF_LINKED)
	{
	    /* Beh. It's linked. Grab the host nick instead. */
	    ni = findnick (ni->host);
	}

	/* Do they have too many? */
	if (lni->linkcnt >= maxlinks)
	{
	    notice (s_NickServ, source, NS_MAX_LINKS_OTHER, lni->nick, maxlinks);
	    return;
	}

	if (!check_ns_auth (s_ChanServ, source, lni))
	    return;

	++lni->linkcnt;
	u->lastnreg = now;
	lni->links = srealloc (lni->links, sizeof (char *) *lni->linkcnt);
	lni->links[lni->linkcnt - 1] = sstrdup (ni->nick);

	/* Now 'merge' the nicks. Any fields that only exist in the source
	   nick are added to the host nick. Any that exist in both are left
	   alone (host takes precidence). We DON'T check the access list.
	   It's too much trouble and I doubt anyone will care.. besides, it's
	   probably a security risk if the nicks have vastly different lists..
	 */
	if (email_on == TRUE && ni->email && !lni->email)
	    lni->email = sstrdup (ni->email);

	if (url_on == TRUE && ni->url && !lni->url)
	    lni->url = sstrdup (ni->url);

	if (uin_on == TRUE && ni->uin && !lni->uin)
	    lni->uin = ni->uin;

	if (sex_on == TRUE && ni->sex && !lni->sex)
	    lni->sex = sstrdup (ni->sex);

	if (age_on == TRUE && ni->age && !lni->age)
	    lni->age = ni->age;

	if (location_on == TRUE && ni->location && !lni->location)
	    lni->location = sstrdup (ni->location);

	if (name_on == TRUE && ni->name && !lni->name)
	    lni->name = sstrdup (ni->name);

	if ((ni->flags & NF_ENFORCE) && !(lni->flags & NF_ENFORCE))
	    lni->flags |= NF_ENFORCE;

	if ((ni->flags & NF_SECURE) && !(lni->flags & NF_SECURE))
	    lni->flags |= NF_SECURE;

	if ((ni->flags & NF_HIDEMAIL) && !(lni->flags & NF_HIDEMAIL))
	    lni->flags |= NF_HIDEMAIL;

	if ((ni->flags & NF_SHOWMAIL) && !(lni->flags & NF_SHOWMAIL))
	    lni->flags |= NF_SHOWMAIL;

	if ((ni->flags & NF_NEVEROP) && !(lni->flags & NF_NEVEROP))
	    lni->flags |= NF_NEVEROP;

	if ((ni->flags & NF_NOOP) && !(lni->flags & NF_NOOP))
	    lni->flags |= NF_NOOP;

	if ((ni->flags & NF_NOSUCCESSOR) && !(lni->flags & NF_NOSUCCESSOR))
	    lni->flags |= NF_NOSUCCESSOR;

	if ((ni->flags & NF_PRIVATE) && !(lni->flags & NF_PRIVATE))
	    lni->flags |= NF_PRIVATE;

	if ((ni->flags & NF_NOTICE) && !(lni->flags & NF_NOTICE))
	    lni->flags |= NF_NOTICE;

	if ((ni->flags & NF_PRIVMSG) && !(lni->flags & NF_PRIVMSG))
	    lni->flags |= NF_PRIVMSG;

	/* Make note of the nickname before we free it */
	strscpy (tnick, ni->nick, NICKLEN);

	/* Done merging .. Drop the source nick. */
	delnick (ni);

	/* Now re-register it, blank, and set it up as a linked nick.
	   We use tni, since the linking nickname doesn't nessecarily
	   have to be source.
	 */
	ni = scalloc (sizeof (NickInfo), 1);
	strscpy (ni->nick, tnick, NICKLEN);
	ni->host = sstrdup (lni->nick);
	ni->flags = NF_LINKED;

	list = &nicklist[NSHASH(ni->nick)];
	ni->next = *list;
	if (*list)
	    (*list)->prev = ni;
	*list = ni;

	notice (s_NickServ, source, NS_NICKS_MERGED, tnick, lni->nick);
	log ("NS:LINK: %s %s %s!%s@%s", tnick, lni->nick,
	    source, u->user, u->host);
	snoop (s_NickServ, "NS:LINK: %s %s %s!%s@%s", tnick,
	    lni->nick, source, u->user, u->host);
    }
}

/* Unlink two nicks. */
static void do_unlink (const char *source)
{
    NickInfo *ni, *lni, **list;
    User *u = finduser (source);
    char *nick = strtok (NULL, " "), **access, tmp[NICKLEN];
    int i;

    if (!u)
	return;

    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "UNLINK NickName");
	return;
    }

    ni = get_nick (u);

    if (!ni)
    {
	notice (s_NickServ, source, ERR_NEED_IDENT, "UNLINK");
	return;
    }

    lni = findnick (nick);

    if (!lni)
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, lni))
	return;

    /* Are they actually linked? */
    if (!lni->host || stricmp (ni->nick, lni->host))
    {
	notice (s_NickServ, source, NS_NOT_LINKED, lni->nick, ni->nick);
	return;
    }

    /* Ok. By this point, we know they're authed for the host nick, and
       we know the given nick is linked to the host nick. Now we seperate
       the two. In most services that I'm aware, unlinking two nicks drops
       the previously linked nick. Not here. When two nicks are unlinked,
       they become standalone nicknames. If someone wants to get rid of a
       linked nick, they can use the drop command, before or after unlinking.

       To seperate two nicknames, we'll delete it, then 'register' it, copying
       all the hosts info into it.
     */
    strscpy (tmp, lni->nick, NICKLEN);

    delnick (lni);

    /* Now fill the fields. */
    lni = scalloc (sizeof (NickInfo), 1);
    strscpy (lni->nick, tmp, NICKLEN);
    strscpy (lni->pass, ni->pass, PASSWDLEN);
    lni->registered = ni->registered;
    lni->lastseen = ni->lastseen;
    lni->timestamp = ni->timestamp;
    lni->sstamp = ni->sstamp;
    lni->memos.memolimit = ni->memos.memolimit;
    lni->flags = ni->flags;
    lni->usermask = sstrdup (ni->usermask);
    lni->real = sstrdup (ni->real);

    if (ni->email)
	lni->email = sstrdup (ni->email);

    if (ni->url)
	lni->url = sstrdup (ni->url);

    if (ni->uin)
	lni->uin = ni->uin;

    if (ni->name)
	lni->name = sstrdup (ni->name);

    if (ni->age)
	lni->age = ni->age;

    if (ni->sex)
	lni->sex = sstrdup (ni->sex);

    if (ni->location)
	lni->location = sstrdup (ni->location);

    if (ni->forward)
	lni->forward = sstrdup (ni->forward);

    for (access = ni->access, i = 0; i < ni->accesscnt; access++, i++)
    {
	++lni->accesscnt;
	lni->access = srealloc (lni->access, sizeof (char *) *lni->accesscnt);
	lni->access[lni->accesscnt - 1] = sstrdup (ni->access[i]);
    }

    list = &nicklist[NSHASH(lni->nick)];
    lni->next = *list;
    if (*list)
	(*list)->prev = lni;
    *list = lni;

    notice (s_NickServ, source, NS_NICKS_UNLINKED, lni->nick, ni->nick);
}

/* List the links for a nickname. */
static void do_links (const char *source)
{
    NickInfo *ni;
    User *u = finduser (source);
    char buf[BUFSIZE], **links, *nick = strtok (NULL, " ");
    int i;

    if (!u)
	return;

    /* If they didn't specify a nick, check if they've identified for a
       nick. If not, then we'll go for their current nick.
     */
    if (nick)
	ni = findnick (nick);
    else
	ni = get_nick (u);

    if (!ni)
    {
	if (nick)
	    notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	else
	    notice (s_NickServ, source, ERR_NEED_IDENT, "LINKS");

	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Go for the host */
    if (ni->host)
	ni = findnick (ni->host);

    *buf = 0;

    /* We store the list of linked nicks for a host on the host
       nick to save cpu time searching for them. Thus, we assume
       the list on the host is always correct (bad things happen
       if its not.)
     */
    for (links = ni->links, i = 0; i < ni->linkcnt; links++, i++)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, *links);
    }

    if (*buf)
    {
	notice (s_NickServ, source, NS_LINKS_LIST, ni->nick, buf);
	return;
    }
    else
    {
 	notice (s_NickServ, source, NS_NO_LINKS, ni->nick);
	return;
    }
}

/* List the channels for a nickname. */
static void do_chans (const char *source)
{
    NickInfo *ni;
    ChanInfo *ci;
    User *u = finduser (source);
    char buf[BUFSIZE], **chans, *nick = strtok (NULL, " ");
    int i;

    if (!u)
	return;

    /* If they didn't specify a nick, check if they've identified for a
       nick. If not, then we'll go for their current nick.
     */
    if (nick)
	ni = findnick (nick);
    else
	ni = get_nick (u);

    if (!ni)
    {
	if (nick)
	    notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	else
	    notice (s_NickServ, source, ERR_NEED_IDENT, "CHANS");

	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Go for the host */
    if (ni->host)
	ni = findnick (ni->host);

    *buf = 0;

    /* We store the list of channels for a host on the host
       nick to save cpu time searching for them. This list may not
       always be correct, though. So we'll check each channel,
       and remove it from the list if it's not registered. This
       is still much faster than checking the founder of every
       channel, though.
     */
    for (chans = ni->chans, i = 0; i < ni->chancnt; chans++, i++)
    {
	ci = cs_findchan (*chans);

	if (!ci)
	{
	    /* This channel isn't registered anymore. Take it out
	       of the chans list, and continue on.
	     */
	    free (*chans);

	    --ni->chancnt;

	    if (i < ni->chancnt)
		bcopy (chans + 1, chans, (ni->chancnt - i) * sizeof (char *));

	    if (ni->chancnt)
		ni->chans = srealloc (ni->chans, ni->chancnt * sizeof (char *));
	    else
	    {
		if (ni->chans)
		    free (ni->chans);

		ni->chans = NULL;
	    }

	    continue;
	}

	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, *chans);
    }

    if (*buf)
    {
	notice (s_NickServ, source, NS_CHANS_LIST, ni->nick, buf);
	return;
    }
    else
    {
 	notice (s_NickServ, source, NS_NO_CHANS, ni->nick);
	return;
    }
}

/* Show the users status to services. */
static void do_status (const char *source)
{
    NickInfo *ni;
    User *u = finduser (source);
    char buf[256];
    int hasstatus = 0;

    if (!u)
	return;

    ni = get_nick (u);

    if (ni)
    {
	hasstatus = 1;

	sprintf (buf, "Identified (%s)", ni->nick);

	if (is_oper (source))
	    strcat (buf, ", IRCOp");

	if (is_csop (source))
	{
	    if (*buf)
		strcat (buf, ", ");

	    strcat (buf, "CSOp");
	}

	if (is_sra (source))
	{
	    if (*buf)
		strcat (buf, ", ");

	    strcat (buf, "SRA");
	}

	notice (s_NickServ, source, NS_STATUS, buf);
    }

    if (!ni)
	ni = findnick (source);

    if (!hasstatus && ni && (ni->status & NS_RECOGNIZED))
    {
	hasstatus = 1;

	strcat (buf, "Recognized");

	if (is_oper (source))
	    strcat (buf, ", IRCOp");

	if (is_csop (source))
	{
	    if (*buf)
		strcat (buf, ", ");

	    strcat (buf, "CSOp");
	}

	if (is_sra (source))
	{
	    if (*buf)
		strcat (buf, ", ");

	    strcat (buf, "SRA");
	}

	notice (s_NickServ, source, NS_STATUS, buf);
    }

    if (u->idchancnt)
    {
	char **chans;
	int i;

	*buf = 0;

	for (chans = u->idchans, i = 0; i < u->idchancnt; chans++, i++)
	{
	    /* If they identified for a channel that was then dropped,
	       take that channel out of their idchans list. This is a
	       potential security problem, since someone could gain access
	       to someone elses channel by having identified for it before
	       it was dropped and reregistered. I'm not too concerned with
	       this for two reasons. First, going through EVERY user, then
	       every channel they've identified for to free them when dropped
	       is likely to be cpu intensive on large nets. Secondly, if
	       someone wants to hack a channel, why would they drop it in the
	       first place? It's much easier to just wait until they log off
	       and its freed then.
	     */
	    if (!cs_findchan (*chans))
	    {
		if (*chans)
		    free (*chans);

		--u->idchancnt;

		if (i < u->idchancnt)
		    bcopy (chans + 1, chans, (u->idchancnt - i) * sizeof (char *));

		if (u->idchancnt)
		    u->idchans = srealloc (u->idchans, u->idchancnt * sizeof (char *));
		else
		{
		    if (u->idchans)
			free (u->idchans);
		    u->idchans = NULL;
		}

		continue;
	    }

	    if (*buf)
		strcat (buf, ", ");

	    strcat (buf, *chans);
	}

	if (*buf)
	{
	    hasstatus = 1;

	    notice (s_NickServ, source, NS_STATUS_CHANS, buf);
	}
    }

    if (!hasstatus)
	notice (s_NickServ, source, NS_STATUS_NONE);
}

/* Show information on a registered nickname. */
static void do_info (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");
    char buf[BUFSIZE];

    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "INFO NickName");
	errmoreinfo (s_NickServ, source, "INFO");
	return;
    }

    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    /* We'll reference hni from here on in */
    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    notice (s_NickServ, source, NS_INFO_START, ni->nick, hni->real);

    if (ni->flags & NF_WAITAUTH)
	notice (s_NickServ, source, NS_INFO_WAITING);

    if (hni->flags & NF_FROZEN)
    {
	notice (s_NickServ, source, NS_INFO_FROZEN);

	notice (s_NickServ, source, NS_INFO_FROZEN_BY, hni->temp);

	if (hni->freezereason)
	    notice (s_NickServ, source, NS_INFO_FROZEN_WHY, hni->freezereason);

	notice (s_NickServ, source, NS_INFO_FROZEN_TIME, get_time (hni->lastseen, 4));
    }

    if (ni->host)
	notice (s_NickServ, source, NS_INFO_HOST, ni->host);

    notice (s_NickServ, source, NS_INFO_REGISTERED,
	get_time (hni->registered, 4), time_ago (hni->registered));

    if (!(hni->flags & NF_PRIVATE) && !finduser (hni->nick))
    {
	notice (s_NickServ, source, NS_INFO_LAST_MASK, hni->usermask);

	notice (s_NickServ, source, NS_INFO_LAST_USED,
	    get_time (hni->lastseen, 4), time_ago (hni->lastseen));
    }

    if (showmail_on == TRUE && email_on == TRUE)
	if (hni->email && ((hni->flags & NF_SHOWMAIL) || is_csop (source))) 
	    notice (s_NickServ, source, NS_INFO_EMAIL, hni->email);

    if (showmail_on == FALSE && email_on == TRUE)
	if (hni->email && (!(hni->flags & NF_HIDEMAIL) || is_csop (source)))
	    notice (s_NickServ, source, NS_INFO_EMAIL, hni->email);

    if (url_on == TRUE)
	if (hni->url)
	    notice (s_NickServ, source, NS_INFO_URL, hni->url);

    if (uin_on == TRUE)
	if (hni->uin)
	    notice (s_NickServ, source, NS_INFO_UIN, hni->uin);

    if (name_on == TRUE)
	if (hni->name)
	    notice (s_NickServ, source, NS_INFO_NAME, hni->name);

    if (age_on == TRUE)
	if (hni->age)
	    notice (s_NickServ, source, NS_INFO_AGE, hni->age);

    if (sex_on == TRUE)
	if (hni->sex)
	    notice (s_NickServ, source, NS_INFO_SEX, hni->sex);

    if (location_on == TRUE)
	if (hni->location)
	    notice (s_NickServ, source, NS_INFO_LOCATION, hni->location);

    if (hni->forward)
	notice (s_NickServ, source, NS_INFO_FORWARD, hni->forward);

    *buf = 0;

    if (hni->flags & NF_ENFORCE)
	strcat (buf, FLAG_ENFORCE);

    if (hni->flags & NF_NEVEROP)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_NEVEROP);
    }

    if (hni->flags & NF_NOOP)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_NOOP);
    }

    if (hni->flags & NF_NOSUCCESSOR)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_NOSUCCESSOR);
    }

    if (hni->flags & NF_PRIVATE)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_PRIVATE);
    }

    if (hni->flags & NF_SECURE)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_SECURE);
    }

    if (hni->flags & NF_MEMOMAIL)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_MEMOMAIL);
    }

    if (hni->flags & NF_RECEIPTS)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_RECEIPTS);
    }

    if (hni->flags & NF_NOMEMO)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_NOMEMO);
    }

    if (hni->flags & NF_HELD)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_HELD);
    }

    if (hni->flags & NF_MARKED)
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_MARKED);
    }

    if (can_csop (hni->nick))
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_CSOP);
    }

    if (can_sra (hni->nick))
    {
	if (*buf)
	    strcat (buf, ", ");

	strcat (buf, FLAG_SRA);
    }

    if (*buf)
	notice (s_NickServ, source, NS_INFO_FLAGS, buf);

    notice (s_NickServ, source, NS_INFO_END);
}

/* Identify to a nickname. */
static void do_identify (const char *source)
{
    NickInfo *ni, *hni;
    User *u = finduser (source);
    struct u_chanlist *c;
    struct c_userlist *cu;
    char *param1 = strtok (NULL, " ");
    char *param2 = strtok (NULL, " ");
    char *pass;

    if (!u)
	return;

    if (!param1)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "IDENTIFY [NickName] Password");
	errmoreinfo (s_NickServ, source, "IDENTIFY");
        return;
    }

    /* Do the password shuffle! */
    if (param2)
    {
	ni = findnick (param1);
	pass = param2;
    }
    else
    {
	ni = findnick (source);
	pass = param1;
    }

    /* Is the nick registered? */
    if (!ni)
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED,
	    param2 ? param1 : source);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* We'll reference hni from here on in */
    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    /* Check the given password against the stored one */
    if (!strcmp (pass, hni->pass))
    {
	/* Passwords match, Update flags, times, and other information */
	hni->lastseen = time (NULL);
	hni->usermask = smalloc (strlen (u->user) + strlen (u->host) + 2);
	hni->timestamp = u->timestamp;
	hni->sstamp = u->sstamp;

	sprintf (hni->usermask, "%s@%s", u->user, u->host);

	if (hni->real)
	    free (hni->real);

	hni->real = sstrdup (u->real);

	if (u->lastnick)
	    free (u->lastnick);

	u->lastnick = sstrdup (hni->nick);

	/* Reset the passfail count. */
	u->passfail = 0;

	/* Set the SRA flag if SecureSRA is off */
	if (securesra_on == FALSE)
	    if (can_sra (hni->nick))
		u->flags |= UF_SRA;

#ifndef IRCUP9
	/* SRA-specific stuff */
	if (is_csop (source))
	    send_cmd (s_NickServ, "SVSMODE %s +ra", source);
	else
	    send_cmd (s_NickServ, "SVSMODE %s +r", source);
#endif

#ifdef UNREAL3
	if (is_sra (source) && securesra_on == FALSE)
	    send_cmd (s_NickServ, "SWHOIS %s :is a Services Root Admin", source);
#endif

	notice (s_NickServ, source, RPL_PASS_ACCEPTED, ni->nick);
	snoop (s_NickServ, "NS:IDENTIFY: %s %s!%s@%s", ni->nick, source, 
	    u->user, u->host);

	/* Check for any new memos */
	check_memos (u);

	/* Check for waiting verifies */
	if (is_csop (source))
	    check_verifies (u);

	/* Tell them about new news */
	if (hni->flags & NF_NEWS)
	{
	    FILE *f;

	    /* Make sure there still IS news to read */
	    f = fopen (NEWSFILE, "r");

	    if (f)
	    {
		notice (s_MemoServ, source, MS_NEW_NEWS);
		notice (s_MemoServ, source, MS_TO_READ_NEWS,
		    haveserv_on == TRUE ? "" : "MSG ", s_MemoServ,
		    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		    haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    me.name : "");
		fclose (f);
	    }
	    else
	    {
		/* The news file is gone now .. Turn the flag off. */
		hni->flags &= ~NF_NEWS;
	    }
	}

	/* Lastly, go through their channels and op them in any they have access
	   in, unless NEVEROP is on. This isn't really NESSECARY, but it's one of
	   those nice little touches that makes everyone happy ;)
	 */
	if (!(ni->flags & NF_NEVEROP))
	{
	    ChanInfo *ci;
	    int ulev = 0;

	    for (c = u->chans; c; c = c->next)
		for (cu = c->chan->users; cu; cu = cu->next)
		    if (!stricmp (cu->user->nick, source))
		    {
			ci = cs_findchan (c->chan->name);

			if (ci)
			    ulev = get_access (u, ci);
			else
			    continue;

			/* VOP? */
			if (ulev == 1 && !(cu->mode & CMODE_v))
			{
			    send_cmode (s_ChanServ, c->chan->name, "+v", u->nick);
			    cu->mode |= CMODE_v;
			}
#ifdef UNREAL3
			/* HOP? */
			if (ulev == 2 && !(cu->mode & CMODE_h))
			{
			    send_cmode (s_ChanServ, c->chan->name, "+h", u->nick);
			    cu->mode |= CMODE_h;
			}
#endif
		 	/* AOP or higher? */
			if (ulev >= 3 && !(cu->mode & CMODE_o))
			{
			    send_cmode (s_ChanServ, c->chan->name, "+o", u->nick);
			    cu->mode |= CMODE_o;
			}
		    }
	}
    }
    else
    {
	/* Passwords don't match, tell them and update their passfail count */
	++u->passfail;

	notice (s_NickServ, source, RPL_WRONG_PASS, ni->nick);

	/* If they've exceeded the limit, kill them. */
	if (u->passfail >= passfails)
	{
	    snoop (s_NickServ,
		"NS:IDENTIFY:BP: %s %s!%s@%s (Killed)",
		ni->nick, source, u->user, u->host);
	    log ("NS:IDENTIFY:BP: %s %s!%s@%s (Killed)",
		ni->nick, source, u->user, u->host);
	    kill_user (s_NickServ, source, RPL_KILLED_PASSFAIL);
	    return;
	}

	if (u->passfail >= (passfails - 1))
	    notice (s_NickServ, source, RPL_LAST_TRY);
	snoop (s_NickServ,
	    "NS:IDENTIFY:BP: %s %s!%s@%s (%d/%d)",
	    ni->nick, source, u->user, u->host, u->passfail, passfails);
	log ("NS:IDENTIFY:BP: %s %s!%s@%s (%d/%d)",
	    ni->nick, source, u->user, u->host, u->passfail, passfails);
    }
}

/* Toggle various NickServ settings */
static void do_set (const char *source)
{
    NickInfo *ni;
    User *u = finduser (source);
    char *option = strtok (NULL, " ");
    char *param = strtok (NULL, "");

    if (!u)
	return;

    /* Do we have all needed info? */
    if (!option || !param)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "SET Option Parameters");
	errmoreinfo (s_NickServ, source, "SET");
	return;
    }

    /* Make sure they're authed for a nick */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_NickServ, source, ERR_NEED_IDENT, "SET");
	return;
    }

    if (ni->flags & NF_FROZEN)
    {
	notice (s_NickServ, source, NS_NICK_FROZEN, ni->nick);
	return;
    }

    /* Toggle ENFORCE? */
    if (!stricmp (option, "ENFORCE"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ni->flags & NF_ENFORCE)
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "Enforce", "on");
		return;
	    }

	    ni->flags |= NF_ENFORCE;
	    notice (s_NickServ, source, NS_OPTION_NOW, "Enforce", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ni->flags & NF_ENFORCE))
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "Enforce", "off");
		return;
	    }

	    ni->flags &= ~NF_ENFORCE;
	    notice (s_NickServ, source, NS_OPTION_NOW, "Enforce", "off");
	    return;
	}

	else
	{
	    notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	}
    }

    /* Toggle SECURE? */
    if (!stricmp (option, "SECURE"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ni->flags & NF_SECURE)
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "Secure", "on");
		return;
	    }

	    ni->flags |= NF_SECURE;
	    notice (s_NickServ, source, NS_OPTION_NOW, "Secure", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ni->flags & NF_SECURE))
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "Secure", "off");
		return;
	    }

	    ni->flags &= ~NF_SECURE;
	    notice (s_NickServ, source, NS_OPTION_NOW, "Secure", "off");
	    return;
	}

	else
	{
	    notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    if (showmail_on == FALSE && email_on == TRUE)
    {
	/* Toggle HIDEMAIL? */
	if (!stricmp (option, "HIDEMAIL"))
	{
	    if (!stricmp (param, "ON"))
	    {
		if (ni->flags & NF_HIDEMAIL)
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "HideMail", "on");
		    return;
		}

		ni->flags &= ~NF_SHOWMAIL;
		ni->flags |= NF_HIDEMAIL;
		notice (s_NickServ, source, NS_OPTION_NOW, "HideMail", "on");
		return;
	    }

	    if (!stricmp (param, "OFF"))
	    {
		if (!(ni->flags & NF_HIDEMAIL))
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "HideMail", "off");
		    return;
		}

		ni->flags &= ~NF_HIDEMAIL;
		notice (s_NickServ, source, NS_OPTION_NOW, "HideMail", "off");
		return;
	    }

	    else
	    {
		notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		    haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		    haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    me.name : "");
		return;
	    }
	}
    }

    if (showmail_on == TRUE && email_on == TRUE)
    {
	/* Toggle SHOWMAIL? */
	if (!stricmp (option, "SHOWMAIL"))
	{
	    if (!stricmp (param, "ON"))
	    {
		if (ni->flags & NF_SHOWMAIL)
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "ShowMail", "on");
		    return;
		}

		ni->flags &= ~NF_HIDEMAIL;
		ni->flags |= NF_SHOWMAIL;
		notice (s_NickServ, source, NS_OPTION_NOW, "ShowMail", "on");
		return;
	    }

	    if (!stricmp (param, "OFF"))
	    {
		if (!(ni->flags & NF_SHOWMAIL))
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "ShowMail", "off");
		    return;
		}

		ni->flags &= ~NF_SHOWMAIL;
		notice (s_NickServ, source, NS_OPTION_NOW, "ShowMail", "off");
		return;
	    }

	    else
	    {
		notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		    haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		    haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    me.name : "");
		return;
	    }
	}
    }

    /* Toggle NEVEROP? */
    if (!stricmp (option, "NEVEROP"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ni->flags & NF_NEVEROP)
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "NeverOp", "on");
		return;
	    }

	    ni->flags |= NF_NEVEROP;
	    notice (s_NickServ, source, NS_OPTION_NOW, "NeverOp", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ni->flags & NF_NEVEROP))
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "NeverOp", "off");
		return;
	    }

	    ni->flags &= ~NF_NEVEROP;
	    notice (s_NickServ, source, NS_OPTION_NOW, "NeverOp", "off");
	    return;
	}

	else
	{
	    notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    /* Toggle NOOP? */
    if (!stricmp (option, "NOOP"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ni->flags & NF_NOOP)
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "NoOp", "on");
		return;
	    }

	    ni->flags |= NF_NOOP;
	    notice (s_NickServ, source, NS_OPTION_NOW, "NoOp", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ni->flags & NF_NOOP))
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "NoOp", "off");
		return;
	    }

	    ni->flags &= ~NF_NOOP;
	    notice (s_NickServ, source, NS_OPTION_NOW, "NoOp", "off");
	    return;
	}

	else
	{
	    notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
 		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    /* Toggle NOSUCCESSOR? */
    if (!stricmp (option, "NOSUCCESSOR"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ni->flags & NF_NOSUCCESSOR)
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "NoSuccessor", "on");
		return;
	    }

	    ni->flags |= NF_NOSUCCESSOR;
	    notice (s_NickServ, source, NS_OPTION_NOW, "NoSuccessor", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ni->flags & NF_NOSUCCESSOR))
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "NoSuccessor", "off");
		return;
	    }

	    ni->flags &= ~NF_NOSUCCESSOR;
	    notice (s_NickServ, source, NS_OPTION_NOW, "NoSuccessor", "off");
	    return;
	}

	else
	{
	    notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    if (privmsg_on == FALSE)
    {
	/* Toggle PRIVMSG? */
	if (!stricmp (option, "PRIVMSG"))
	{
	    if (!stricmp (param, "ON"))
	    {
		if (ni->flags & NF_PRIVMSG)
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "PrivMSG", "on");
		    return;
		}

		ni->flags &= ~NF_NOTICE;
		ni->flags |= NF_PRIVMSG;
		notice (s_NickServ, source, NS_OPTION_NOW, "PrivMSG", "on");
		return;
	    }

	    if (!stricmp (param, "OFF"))
	    {
		if (!(ni->flags & NF_PRIVMSG))
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "PrivMSG", "off");
		    return;
		}

		ni->flags &= ~NF_PRIVMSG;
		notice (s_NickServ, source, NS_OPTION_NOW, "PrivMSG", "off");
		return;
	    }

	    else
	    {
		notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		    haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		    haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    me.name : "");
		return;
	    }
	}
    }

    if (privmsg_on == TRUE)
    {
	/* Toggle NOTICE? */
	if (!stricmp (option, "NOTICE"))
	{
	    if (!stricmp (param, "ON"))
	    {
		if (ni->flags & NF_NOTICE)
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "Notice", "on");
		    return;
		}

		ni->flags &= ~NF_PRIVMSG;
		ni->flags |= NF_NOTICE;
		notice (s_NickServ, source, NS_OPTION_NOW, "Notice", "on");
		return;
	    }

	    if (!stricmp (param, "OFF"))
	    {
		if (!(ni->flags & NF_NOTICE))
		{
		    notice (s_NickServ, source, NS_OPTION_ALREADY, "Notice", "off");
		    return;
		}

		ni->flags &= ~NF_NOTICE;
		notice (s_NickServ, source, NS_OPTION_NOW, "Notice", "off");
		return;
	    }

	    else
	    {
		notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		    haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		    haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		    haveserv_on == TRUE ? "" : securitysetting == 1 ?
		    me.name : "");
		return;
	    }
	}
    }

    /* Toggle PRIVATE? */
    if (!stricmp (option, "PRIVATE"))
    {
	if (!stricmp (param, "ON"))
	{
	    if (ni->flags & NF_PRIVATE)
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "Private", "on");
		return;
	    }

	    ni->flags |= NF_PRIVATE;
	    notice (s_NickServ, source, NS_OPTION_NOW, "Private", "on");
	    return;
	}

	if (!stricmp (param, "OFF"))
	{
	    if (!(ni->flags & NF_PRIVATE))
	    {
		notice (s_NickServ, source, NS_OPTION_ALREADY, "Private", "off");
		return;
	    }

	    ni->flags &= ~NF_PRIVATE;
	    notice (s_NickServ, source, NS_OPTION_NOW, "Private", "off");
	    return;
	}

	else
	{
	    notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (param),
		haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
		haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
		haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
	    return;
	}
    }

    /* Change password? */
    if (!stricmp (option, "PASS") || !stricmp (option, "PASSWD") ||
	!stricmp (option, "PASSWORD"))
    {
	int reason = 0;
	char *pass1 = strtok (param, " ");
	char *pass2 = strtok (NULL, " ");

	if (!pass2)
	{
	    notice (s_NickServ, source, RPL_SYNTAX,
		"SET PASS Password Password");
	    errmoreinfo (s_NickServ, source, "SET PASS");
	    return;
	}

	/* Password sanity checks */
	if (strlen (pass1) < 5)
	    reason = 1;
	if (strlen (pass1) > PASSWDLEN)
	    reason = 2;
	if (!stricmp (source, pass1))
	    reason = 3;
	if (!stricmp ("password", pass1))
	    reason = 4;
	if (strcmp (pass1, pass2))
	    reason = 5;

	if (reason)
	{
	    notice (s_NickServ, source, ERR_BETTER_PASS);

	    if (reason == 1)
		notice (s_NickServ, source, ERR_PASS_REJECTED_SHORT);
	    if (reason == 2)
		notice (s_NickServ, source, ERR_PASS_REJECTED_LONG);
	    if (reason == 3)
		notice (s_NickServ, source, ERR_PASS_REJECTED_NICK);
	    if (reason == 4)
		notice (s_NickServ, source, ERR_PASS_REJECTED_WEAK);
	    if (reason == 5)
		notice (s_ChanServ, source, ERR_PASS_REJECTED_DIFF);

	    errmoreinfo (s_NickServ, source, "SET PASS");
	    return;
	}

	/* Passed checks, set it */
	strscpy (ni->pass, pass1, PASSWDLEN);
	notice (s_NickServ, source, RPL_PASSWORD_SET, ni->pass);

	/* Log this. This might be an attempt to steal a nick.. */
	log ("NS:SET:PASS: %s %s!%s@%s", ni->nick, source, u->user, u->host);
	return;
    }

    /* Change E-Mail? */
    if (email_on == TRUE)
    {
	if (!stricmp (option, "EMAIL"))
	{
	    if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	    {
		if (ni->email)
		    free (ni->email);

		ni->email = NULL;

		notice (s_NickServ, source, RPL_REMOVED, "Your \2E-Mail\2");
		return;
	    }
	    else
	    {
		/* Check if we're waiting for confirmation on an E-Mail change. */
		if (ni->temp && ni->key && (nsregistertype == 1 || nsregistertype == 3 ||
		    nsregistertype == 6))
		{
		    if (atoi (param) == ni->key)
		    {
			free (ni->email);
			ni->email = sstrdup (ni->temp);
			free (ni->temp);
			ni->key = 0;

			notice (s_NickServ, source, RPL_ADDED, "Your \2E-Mail\2", ni->email);

			return;
		    }
		}

		if (validemail (param, source))
		{
		    /* If we're using E-Mail AUTH register types, we'll confirm
		       this new address with an email. If we get the right code,
		       we'll change their E-Mail.
		    */
		    if (nsregistertype == 1 || nsregistertype == 3 ||
			nsregistertype == 6)
		    {
			/* Is this email being abused? */
			if (check_email (source, param))
			{
			    notice (s_NickServ, source, ERR_EMAIL_USED);
			    return;
			}

			/* Make a key */
			ni->key = makekey ();

			/* Put the E-Mail they want into temp */
			ni->temp = sstrdup (param);

			/* Send the email */
			sendemail (ni->nick, itoa (ni->key), 6);

			/* Tell them to expect an email .. */
			notice (s_NickServ, source, NS_CONFIRM_EMAIL, ni->temp);
			notice (s_NickServ, source, NS_CONFIRM_CHANGE);
			notice (s_NickServ, source, NS_CONFIRM_TIMEOUT,
			    duration (tempexpire, 2));
		    }
		    else
		    {
			if (ni->email)
			    free (ni->email);

			ni->email = sstrdup (param);

			notice (s_NickServ, source, RPL_ADDED, "Your \2E-Mail\2", param);
		    }
		}
		else
		    notice (s_NickServ, source, ERR_INVALID, "E-Mail", "");

		return;
	    }
	}
    }

    /* Change URL? */
    if (url_on == TRUE)
    {
	if (!stricmp (option, "URL"))
	{
	    if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	    {
		if (ni->url)
		    free (ni->url);
		ni->url = NULL;
		notice (s_NickServ, source, RPL_REMOVED, "Your \2URL\2");
		return;
	    }
	    else
	    {
		if (strstr (param, "://"))
		{
		    if (ni->url)
			free (ni->url);
		    ni->url = sstrdup (param);
		    notice (s_NickServ, source, RPL_ADDED, "Your \2URL\2", param);
		}
		else
		    notice (s_NickServ, source, ERR_INVALID, "URL", "");

		return;
	    }
	}
    }

    /* Change UIN? */
    if (uin_on == TRUE)
    {
	if (!stricmp (option, "UIN"))
	{
	    if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	    {
		ni->uin = 0;
		notice (s_NickServ, source, RPL_REMOVED, "Your \2UIN\2");
		return;
	    }
	    else
	    {
		if (atoi (param) > 1001 && atoi (param) < 2147483646)
		{
		    ni->uin = atoi (param);
		    notice (s_NickServ, source, RPL_ADDED, "Your \2UIN\2", param);
		}
		else
		    notice (s_NickServ, source, ERR_INVALID, "UIN", "");

		return;
	    }
	}
    }

    /* Change Name? */
    if (name_on == TRUE)
    {
	/* Change NAME? */
	if (!stricmp (option, "NAME"))
	{
	    if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	    {
		free (ni->name);
		ni->name = NULL;
		notice (s_NickServ, source, RPL_REMOVED, "Your \2Name\2");
		return;
	    }
	    else
	    {
		if (strlen (param) > 100)
		    notice (s_NickServ, source, ERR_LENGTH, "name",
			atoi ("100"));
		else
		{
		    free (ni->name);
		    ni->name = sstrdup (param);
		    notice (s_NickServ, source, RPL_ADDED, "Your \2Name\2",
			param);
		}

		return;
	    }
	}
    }

    /* Change Location? */
    if (location_on == TRUE)
    {
	/* Change LOCATION? */
	if (!stricmp (option, "LOCATION"))
	{
	    if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	    {
		free (ni->location);
		ni->location = NULL;
		notice (s_NickServ, source, RPL_REMOVED, "Your \2Location\2");
		return;
	    }
	    else
	    {
		if (strlen (param) > 100)
		    notice (s_NickServ, source, ERR_LENGTH, "location",
			atoi ("100"));
		else
		{
		    free (ni->location);
		    ni->location = sstrdup (param);
		    notice (s_NickServ, source, RPL_ADDED, "Your \2Location\2",
			param);
		}

		return;
	    }
	}
    }

    /* Change Sex? (No comment ..) */
    if (sex_on == TRUE)
    {
	/* Change SEX? */
	if (!stricmp (option, "SEX"))
	{
	    if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	    {
		free (ni->sex);
		ni->sex = NULL;
		notice (s_NickServ, source, RPL_REMOVED, "Your \2Sex\2");
		return;
	    }

	    else if (!stricmp (param, "MALE"))
	    {
		free (ni->sex);
		ni->sex = sstrdup ("Male");
		notice (s_NickServ, source, RPL_ADDED, "Your \2Sex\2", "Male");
		return;
	    }

	    else if (!stricmp (param, "FEMALE"))
	    {
		free (ni->sex);
		ni->sex = sstrdup ("Female");
		notice (s_NickServ, source, RPL_ADDED, "Your \2Sex\2",
		    "Female");
		return;
	    }
	    else
	    {
		notice (s_NickServ, source, ERR_INVALID, "Sex",
		    " SEX must be either MALE or FEMALE.");
		return;
	    }
	}
    }

    /* Change Age? */
    if (age_on == TRUE)
    {
	/* Change AGE? */
	if (!stricmp (option, "AGE"))
	{
	    if (!stricmp (param, "NONE") || !stricmp (param, "OFF"))
	    {
		ni->age = 0;
		notice (s_NickServ, source, RPL_REMOVED, "Your \2Age\2");
		return;
	    }
	    else
	    {
		if (atoi (param) >= 5 && atoi (param) <= 100)
		{
		    ni->age = atoi (param);
		    notice (s_NickServ, source, RPL_ADDED, "Your \2Age\2",
			param);
		}
		else
		    notice (s_NickServ, source, ERR_INVALID, "Age", "");

		return;
	    }
	}
    }

    /* Don't know what they want .. */
    notice (s_NickServ, source, ERR_UNKNOWN_OPTION, strupper (option),
	haveserv_on == TRUE ? "" : "MSG ", s_NickServ,
	haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");
}

/* Add/remove/list access list entries */
static void do_access (const char *source)
{
    NickInfo *ni;
    User *u = finduser (source);
    char **access;
    int i;
    char *option = strtok (NULL, " ");
    char *param = strtok (NULL, "");

    if (!u)
	return;

    /* Do we have all needed info? */
    if (!option)
    {
	notice (s_NickServ, source, RPL_SYNTAX,
	    "ACCESS ADD|DEL|LIST|CURRENT [Mask]");
	errmoreinfo (s_NickServ, source, "ACCESS");
	return;
    }

    /* Make sure they're authed for a nick */
    ni = get_nick (u);

    if (!ni)
    {
	notice (s_NickServ, source, ERR_NEED_IDENT, "ACCESS");
	return;
    }

    /* Add? */
    if (!stricmp (option, "ADD"))
    {
	if (!param)
	{
	    notice (s_NickServ, source, RPL_SYNTAX, "ACCESS ADD Mask");
	    errmoreinfo (s_NickServ, source, "ACCESS");
	    return;
	}

	if (!strchr (param, '@'))
	{
	    notice (s_NickServ, source, ERR_FORMATTING, "user@host");
	    errmoreinfo (s_NickServ, source, "ACCESS");
	    return;
	}

	for (access = ni->access, i = 0; i < ni->accesscnt; access++, i++)
	{
	    if (match_wild_nocase (param, *access))
	    {
		notice (s_NickServ, source, ERR_LIST_THERE, param,
		    "your", "\2ACCESS\2");
		return;
	    }
	}

	++ni->accesscnt;
	ni->access = srealloc (ni->access, sizeof (char *) *ni->accesscnt);
	ni->access[ni->accesscnt - 1] = sstrdup (param);

	notice (s_NickServ, source, NS_ACCESS_ADDED, param);
    }

    /* Del? */
    if (!stricmp (option, "DEL"))
    {
	if (!param)
	{
	    notice (s_NickServ, source, RPL_SYNTAX, "ACCESS DEL Mask|Num");
	    errmoreinfo (s_NickServ, source, "ACCESS");
	    return;
	}

	if (strspn (param, "1234567890") == strlen (param) &&
	    (i = atoi (param)) > 0 && i <= ni->accesscnt)
	{
	    --i;
	    access = &ni->access[i];
	}
	else
	{
	    for (access = ni->access, i = 0; i < ni->accesscnt; ++access, ++i)
		if (!strcmp (*access, param))
		    break;

	    if (i == ni->accesscnt)
		for (access = ni->access, i = 0; i < ni->accesscnt;
		    ++access, ++i)
		    if (!stricmp (*access, param))
			break;

	    if (i == ni->accesscnt)
            {
		notice (s_NickServ, source, ERR_LIST_NOT_THERE, param,
		    "your \2ACCESS\2");
		return;
	    }
	}

	notice (s_NickServ, source, NS_ACCESS_REMOVED, *access);

	if (*access)
	    free (*access);

	--ni->accesscnt;

	if (i < ni->accesscnt)
	    bcopy (access + 1, access, (ni->accesscnt - i) * sizeof (char *));

	if (ni->accesscnt)
	    ni->access = srealloc (ni->access, ni->accesscnt * sizeof (char *));
	else
	{
	    if (ni->access)
		free (ni->access);
	    ni->access = NULL;
	}
    }

    /* Current? */
    if (!stricmp (option, "CURRENT"))
    {
	char *current = create_mask (finduser (source));

	for (access = ni->access, i = 0; i < ni->accesscnt; access++, i++)
	{
	    if (match_wild_nocase (current, *access))
	    {
		notice (s_NickServ, source, ERR_LIST_THERE, current,
		    "your", "\2ACCESS\2");
		return;
	    }
	}

	++ni->accesscnt;
	ni->access = srealloc (ni->access, sizeof (char *) *ni->accesscnt);
	ni->access[ni->accesscnt - 1] = sstrdup (current);
	notice (s_NickServ, source, NS_ACCESS_ADDED, current);
    }

    /* List? */
    if (!stricmp (option, "LIST"))
    {
	notice (s_NickServ, source, NS_ACCESS_LIST, ni->nick);
	for (access = ni->access, i = 0; i < ni->accesscnt; access++, i++)
	{
	    if (param && !match_wild_nocase (param, *access))
		continue;

	    notice (s_NickServ, source, "%-4d %s", i + 1, *access);
	}
	notice (s_NickServ, source, NS_ACCESS_END, i);
    }
}

/* Recover a nickname. */
static void do_recover (const char *source)
{
    NickInfo *ni, *hni;
    User *u, *su = finduser (source);
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");

    if (!su)
	return;

    /* Do we have all needed info? */
    if (!nick || !pass)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "RECOVER Nickname Password");
	errmoreinfo (s_NickServ, source, "RECOVER");
	return;
    }

    /* Don't let them do it on themselves. */
    if (!stricmp (nick, source))
    {
	notice (s_NickServ, source, NS_NOT_SELF, "RECOVER");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    /* Check the passwords. */
    if (strcmp (pass, hni->pass))
    {
	/* Passwords don't match, tell them and update their passfail count */
	++su->passfail;

	notice (s_NickServ, source, RPL_WRONG_PASS, ni->nick);

	/* If they've exceeded the limit, kill them. */
	if (su->passfail >= passfails)
	{
	    snoop (s_NickServ, "NS:RECOVER:BP: %s %s!%s@%s (Killed)",
		ni->nick, source, su->user, su->host);
	    log ("NS:RECOVER:BP: %s %s!%s@%s (Killed)",
		ni->nick, source, su->user, su->host);
	    kill_user (s_NickServ, source, RPL_KILLED_PASSFAIL);
	    return;
	}

	if (su->passfail >= (passfails - 1))
	    notice (s_NickServ, source, RPL_LAST_TRY);

	snoop (s_NickServ, "NS:RECOVER:BP: %s %s!%s@%s (%d/%d)",
            ni->nick, source, su->user, su->host, su->passfail, passfails);
        log ("NS:RECOVER:BP: %s %s!%s@%s (%d/%d)",
	    ni->nick, source, su->user, su->host, su->passfail, passfails);

	return;
    }

    /* Passwords check out, now either remove an enforcer or kill a user
       using the nick. */
    if (ni->status & NS_ENFORCED)
    {
	release (ni, 0);

	snoop (s_NickServ, "NS:RECOVER: %s %s!%s@%s", ni->nick, source,
	    su->user, su->host);

	notice (s_NickServ, source, NS_RECOVERED, ni->nick);
	return;
    }

    /* Not an enforcer.. is it a user thats online? */
    if ((u = finduser (nick)))
    {
	collide (ni, 0);

	snoop (s_NickServ, "NS:RECOVER: %s %s!%s@%s", ni->nick, source,
	    su->user, su->host);

	notice (s_NickServ, source, NS_RECOVERED, ni->nick);
	return;
    }
 
    /* If we get here, we found neither an enforcer or a user. Oh well. */
    notice (s_NickServ, source, NS_NOT_ONLINE, ni->nick);
}

/* Retrieve the password for a nickname */
static void do_getpass (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "GETPASS Nickname");
	errmoreinfo (s_NickServ, source, "GETPASS");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    /* Check if it's marked */
    if (hni->flags & NF_MARKED)
    {
	notice (s_NickServ, source, NS_NICK_MARKED, hni->nick, "GET");
	return;
    }

    notice (s_NickServ, source, RPL_GETPASS, ni->nick, ni->host ? " (" : "",
	ni->host ? ni->host : "", ni->host ? ")" : "", ni->host ? hni->pass :
	ni->pass);

    globops (s_NickServ, RPL_PASS_GLOBAL, source, "GETPASS", ni->nick,
	ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");

    snoop (s_NickServ, "NS:GETPASS: %s %s%s%s%s", source, ni->nick,
	ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");

    log ("NS:GETPASS: %s %s%s%s%s", source, ni->nick,
        ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");
}

/* Set the password for a nickname */
static void do_setpass (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");
    char *pass1 = strtok (NULL, " ");
    char *pass2 = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick || !pass1 || !pass2)
    {
	notice (s_NickServ, source, RPL_SYNTAX,
	    "SETPASS Nickname NewPass NewPass");
	errmoreinfo (s_NickServ, source, "SETPASS");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    if (strcmp (pass1, pass2))
    {
	notice (s_NickServ, source, ERR_PASS_DONT_MATCH);
	return;
    }

    if (ni->host)
    {
	hni = findnick (ni->host);

	/* Check if it's marked */
	if (hni->flags & NF_MARKED)
	{
	    notice (s_NickServ, source, NS_NICK_MARKED, hni->nick, "SET");
	    return;
 	}

	strscpy (hni->pass, pass1, sizeof (hni->pass));
    }
    else
    {
	hni = ni;

	/* Check if it's marked */
	if (hni->flags & NF_MARKED)
	{
	    notice (s_NickServ, source, NS_NICK_MARKED, hni->nick, "SET");
	    return;
	}

	strscpy (ni->pass, pass1, sizeof (ni->pass));
    }

    notice (s_NickServ, source, RPL_SETPASS, ni->nick, ni->host ? " (" : "",
	ni->host ? ni->host : "", ni->host ? ")" : "", ni->host ? hni->pass :
	ni->pass);

    globops (s_NickServ, RPL_PASS_GLOBAL, source, "SETPASS", ni->nick,
	ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");

    snoop (s_NickServ, "NS:SETPASS: %s %s%s%s%s", source, ni->nick,
	ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");

    log ("NS:SETPASS: %s %s%s%s%s", source, ni->nick,
        ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");
}

/* Email the password for a nickname */
static void do_sendpass (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "SENDPASS Nickname");
	errmoreinfo (s_NickServ, source, "SENDPASS");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    if (ni->host)
    {
	hni = findnick (ni->host);

	/* Check if it's marked */
	if (hni->flags & NF_MARKED)
	{
	    notice (s_NickServ, source, NS_NICK_MARKED, hni->nick, "SEND");
	    return;
	}

	if (!hni->email)
	{
	    notice (s_NickServ, source, NS_CANT_SENDPASS, hni->nick);
	    return;
	}
    }
    else
    {
	hni = ni;

	/* Check if it's marked */
	if (hni->flags & NF_MARKED)
	{
	    notice (s_NickServ, source, NS_NICK_MARKED, hni->nick, "SEND");
	    return;
	}

	if (!ni->email)
	{
	    notice (s_NickServ, source, NS_CANT_SENDPASS, ni->nick);
	    return;
	}
    }

    notice (s_NickServ, source, RPL_SENDPASS, ni->nick, ni->host ? " (" : "",
	ni->host ? ni->host : "", ni->host ? ")" : "", ni->host ? hni->email :
	ni->email);

    globops (s_NickServ, RPL_PASS_GLOBAL, source, "SENDPASS", ni->nick,
	ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");

    snoop (s_NickServ, "NS:SENDPASS: %s %s%s%s%s", source, ni->nick,
	ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");

    log ("NS:SENDPASS: %s %s%s%s%s", source, ni->nick,
        ni->host ? " (" : "", ni->host ? hni->nick : "", ni->host ? ")" : "");

    sendemail (ni->host ? hni->nick : ni->nick, ni->email, 2);
}

/* Freeze a nickname */
static void do_freeze (const char *source)
{
    NickInfo *ni, *hni, *tni;
    User *u;
    char *nick = strtok (NULL, " ");
    char *reason = strtok (NULL, "");
    char **links;
    int i;

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "FREEZE NickName [Reason]");
	errmoreinfo (s_NickServ, source, "FREEZE");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Freeze is effective on the host nick, not just one linked nick. */
    if (ni->host)
	hni = findnick (ni->host);
    else
	hni = ni;

    /* Is it already frozen? */
    if (hni->flags & NF_FROZEN)
    {
	notice (s_NickServ, source, NS_ALREADY_FROZEN, hni->nick);
	return;
    }

    hni->flags |= NF_FROZEN;

    /* We'll use the lastseen timestamp for a freeze time. */
    hni->lastseen = time (NULL);

    /* Use temp field for freeze setter. This nick won't be used
       when frozen, so this is safe.
     */
    if (hni->temp)
	free (hni->temp);

    hni->temp = sstrdup (source);

    if (reason)
    {
	if (hni->freezereason)
	    free (hni->freezereason);

	hni->freezereason = sstrdup (reason);
    }

    /* See if this nick is in use. Deal with it if it is. */
    u = finduser (hni->nick);

    if (u)
    {
	notice (s_NickServ, u->nick, NS_INFO_FROZEN);

	if (hni->freezereason)
	    notice (s_NickServ, u->nick, NS_FROZEN_REASON, hni->freezereason);

	if (collidetype != 6)
	    notice (s_NickServ, u->nick, NS_FROZEN_CHANGED);
	else
	    notice (s_NickServ, u->nick, NS_FROZEN_KILLED);

	collide (hni, 0);
    }

    /* Go through the nicks linked to this host and deal with any
       which are in use.
     */
    for (links = hni->links, i = 0; i < hni->linkcnt; links++, i++)
    {
	u = finduser (*links);

	if (u)
	{
	    notice (s_NickServ, u->nick, NS_INFO_FROZEN);

	    if (hni->freezereason)
		notice (s_NickServ, u->nick, NS_FROZEN_REASON, hni->freezereason);

	    if (collidetype != 6)
		notice (s_NickServ, u->nick, NS_FROZEN_CHANGED);
	    else
		notice (s_NickServ, u->nick, NS_FROZEN_KILLED);

	    tni = findnick (u->nick);

	    collide (tni, 0);
	}
    }

    notice (s_NickServ, source, NS_FROZEN, hni->nick);
    log ("NS:FREEZE: %s %s", source, hni->nick);
    snoop (s_NickServ, "NS:FREEZE: %s froze %s", source, hni->nick);
}

/* Undo a freeze on a nickname */
static void do_unfreeze (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "UNFREEZE NickName");
	errmoreinfo (s_NickServ, source, "UNFREEZE");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Freeze is effective on the host nick, not just one linked nick. */
    if (ni->host)
    {
	hni = findnick (ni->host);

	/* Is it frozen? */
	if (!(hni->flags & NF_FROZEN))
	{
	    notice (s_NickServ, source, NS_NOT_FROZEN, hni->nick);
	    return;
	}

	hni->flags &= ~NF_FROZEN;

	free (hni->temp);

	hni->temp = NULL;

	/* Set the last used timestamp to now. We don't want nicks expiring because
	   they've been frozen for more than expiretime..
	 */
	hni->lastseen = time (NULL);

	if (hni->freezereason)
	{
	    free (hni->freezereason);

	    hni->freezereason = NULL;
	}

	notice (s_NickServ, source, NS_UNFROZEN_HOST, ni->nick, hni->nick);
	log ("NS:UNFREEZE: %s %s (%s)", source, hni->nick, ni->nick);
	snoop (s_NickServ, "NS:UNFREEZE: %s unfroze %s (%s)", source, hni->nick, ni->nick);
    }
    else
    {
	/* Is it frozen? */
	if (!(ni->flags & NF_FROZEN))
	{
	    notice (s_NickServ, source, NS_NOT_FROZEN, ni->nick);
	    return;
	}

	ni->flags &= ~NF_FROZEN;

	free (ni->temp);

	ni->temp = NULL;

	/* Set the last used timestamp to now. We don't want nicks expiring because
	   they've been frozen for more than expiretime..
	 */
	ni->lastseen = time (NULL);

	if (ni->freezereason)
	{
	    free (ni->freezereason);

	    ni->freezereason = NULL;
	}

	notice (s_NickServ, source, NS_UNFROZEN, ni->nick);
	log ("NS:UNFREEZE: %s %s", source, ni->nick);
	snoop (s_NickServ, "NS:UNFREEZE: %s unfroze %s", source, ni->nick);
    }
}

/* Hold a nickname */
static void do_hold (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "HOLD NickName");
	errmoreinfo (s_NickServ, source, "HOLD");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Hold is effective on the host nick, not just one linked nick. */
    if (ni->host)
    {
	hni = findnick (ni->host);

	hni->flags |= NF_HELD;

	notice (s_NickServ, source, NS_HELD_HOST, ni->nick, hni->nick);
	log ("NS:HOLD: %s %s (%s)", source, hni->nick, ni->nick);
	snoop (s_NickServ, "NS:HOLD: %s held %s (%s)", source, hni->nick, ni->nick);
    }
    else
    {
	ni->flags |= NF_HELD;

	notice (s_NickServ, source, NS_HELD, ni->nick);
	log ("NS:HOLD: %s %s", source, ni->nick);
	snoop (s_NickServ, "NS:HOLD: %s held %s", source, ni->nick);
    }
}

/* Undo a hold on a nickname */
static void do_unhold (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "UNHOLD NickName");
	errmoreinfo (s_NickServ, source, "UNHOLD");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Hold is effective on the host nick, not just one linked nick. */
    if (ni->host)
    {
	hni = findnick (ni->host);

	hni->flags &= ~NF_HELD;

	notice (s_NickServ, source, NS_UNHELD_HOST, ni->nick, hni->nick);
	log ("NS:UNHOLD: %s %s (%s)", source, hni->nick, ni->nick);
	snoop (s_NickServ, "NS:UNHOLD: %s unheld %s (%s)", source, hni->nick, ni->nick);
    }
    else
    {
	ni->flags &= ~NF_HELD;

	notice (s_NickServ, source, NS_UNHELD, ni->nick);
	log ("NS:UNHOLD: %s %s", source, ni->nick);
	snoop (s_NickServ, "NS:UNHOLD: %s unheld %s", source, ni->nick);
    }
}

/* Mark a nickname */
static void do_mark (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "MARK NickName");
	errmoreinfo (s_NickServ, source, "MARK");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Mark is effective on the host nick, not just one linked nick. */
    if (ni->host)
    {
	hni = findnick (ni->host);

	hni->flags |= NF_MARKED;

	notice (s_NickServ, source, NS_MARKED_HOST, ni->nick, hni->nick);
	log ("NS:MARKED: %s %s (%s)", source, hni->nick, ni->nick);
	snoop (s_NickServ, "NS:MARK: %s marked %s (%s)", source, hni->nick, ni->nick);
    }
    else
    {
	ni->flags |= NF_MARKED;

	notice (s_NickServ, source, NS_MARKED, ni->nick);
	log ("NS:MARK: %s %s", source, ni->nick);
	snoop (s_NickServ, "NS:MARK: %s marked %s", source, ni->nick);
    }
}

/* Undo a mark on a nickname */
static void do_unmark (const char *source)
{
    NickInfo *ni, *hni;
    char *nick = strtok (NULL, " ");

    /* Do we have all needed info? */
    if (!nick)
    {
	notice (s_NickServ, source, RPL_SYNTAX, "UNMARK NickName");
	errmoreinfo (s_NickServ, source, "UNMARK");
	return;
    }

    /* Is the nick registered? */
    if (!(ni = findnick (nick)))
    {
	notice (s_NickServ, source, NS_NOT_REGISTERED, nick);
	return;
    }

    if (!check_ns_auth (s_NickServ, source, ni))
	return;

    /* Mark is effective on the host nick, not just one linked nick. */
    if (ni->host)
    {
	hni = findnick (ni->host);

	hni->flags &= ~NF_MARKED;

	notice (s_NickServ, source, NS_UNMARKED_HOST, ni->nick, hni->nick);
	log ("NS:UNMARK: %s %s (%s)", source, hni->nick, ni->nick);
	snoop (s_NickServ, "NS:UNMARK: %s unmarked %s (%s)", source, hni->nick, ni->nick);
    }
    else
    {
	ni->flags &= ~NF_MARKED;

	notice (s_NickServ, source, NS_UNMARKED, ni->nick);
	log ("NS:UNMARK: %s %s", source, ni->nick);
	snoop (s_NickServ, "NS:UNMARK: %s unmarked %s", source, ni->nick);
    }
}

/* Load the NickServ DB from disk */
void load_ns_dbase (void)
{
    FILE *f = fopen (NICKSERV_DB, "r");
    NickInfo *ni = NULL;
    MemoInfo *mi;
    Memo *m;
    char *item, dBuf[2048], backup [2048];
    int log = 0, nin = 0, min = 0, lin = 0;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

#ifdef HAVE_GETTIMEOFDAY
    log_sameline (1, "NS:DB: Loading... ");
#else
    log ("NS:DB: Loading... ");
#endif

    if (!f)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "No %s found! ", NICKSERV_DB);
#else
	log ("NS:DB: No %s found!", NICKSERV_DB);
#endif
	log = 1;

	/* Check for a db.save */
	strcpy (backup, NICKSERV_DB);
	strcat (backup, ".save");

	if (rename (backup, NICKSERV_DB) > 0)
	{
#ifdef HAVE_GETTIMEOFDAY
	    log_sameline (0, "Recovering %s.save", NICKSERV_DB);
#else
	    log ("NS:DB: Recovering %s.save", NICKSERV_DB);
#endif
	    f = fopen (NICKSERV_DB, "r");
	}
	else
	    return;
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");

	/* Ignore certain things.. */
	if (*item == '#' || *item == '\n' || *item == '\t' || *item == ' ' ||
	    *item == '\0' || *item == '\r')
	    continue;

	/* Services Stamp */
	if (!strcmp (item, "SS"))
	{
	    /* Add 1000 to the stamp each time we load. This helps to
	       prevent duplicate stamps if theres any rollback on the
	       dbs.
	     */
	    sstamp = atoi (strtok (NULL, "\n")) + 1000;
	    continue;
	}

	/* NickInfo */
	if (!strcmp (item, "NI"))
	{
	    NickInfo **list;

	    ni = scalloc (sizeof (NickInfo), 1);
	    mi = &ni->memos;

	    strscpy (ni->nick, strtok (NULL, " "), NICKLEN);
	    strscpy (ni->pass, strtok (NULL, " "), PASSWDLEN);
	    ni->registered = atol (strtok (NULL, " "));
	    ni->lastseen = atol (strtok (NULL, " "));
	    ni->timestamp = atol (strtok (NULL, " "));
	    ni->sstamp = atol (strtok (NULL, " "));
	    ni->flags = atol (strtok (NULL, " "));
            mi->memolimit = atoi (strtok (NULL, " "));
	    ni->key = atoi (strtok (NULL, " "));
	    ni->usermask = sstrdup (strtok (NULL, " "));
	    ni->real = sstrdup (strtok (NULL, "\n"));

	    list = &nicklist[NSHASH(ni->nick)];
	    ni->next = *list;
	    if (*list)
		(*list)->prev = ni;
	    *list = ni;

	    nin++;

	    continue;
	}

	/* Freeze Reason */
	if (!strcmp (item, "FR"))
	{
	    ni->freezereason = sstrdup (strtok (NULL, "\n"));
	    continue;
	}

	/* E-Mail */
	if (email_on == TRUE)
	    if (!strcmp (item, "EM"))
	    {
		ni->email = sstrdup (strtok (NULL, "\n"));
		continue;
	    }

	/* Temp E-Mail */
	if (email_on == TRUE)
	    if (!strcmp (item, "TE"))
	    {
		ni->temp = sstrdup (strtok (NULL, "\n"));
		continue;
	    }

	/* URL */
	if (url_on == TRUE)
	    if (!strcmp (item, "UR"))
	    {
		ni->url = sstrdup (strtok (NULL, "\n"));
		continue;
	    }

	/* UIN */
	if (uin_on == TRUE)
	    if (!strcmp (item, "UN"))
	    {
		ni->uin = atoi (strtok (NULL, "\n"));
		continue;
	    }

	/* Name */
	if (name_on == TRUE)
	    if (!strcmp (item, "NA"))
	    {
		ni->name = sstrdup (strtok (NULL, "\n"));
		continue;
	    }

	/* Age */
	if (age_on == TRUE)
	    if (!strcmp (item, "AG"))
	    {
		ni->age = atoi (strtok (NULL, "\n"));
		continue;
	    }

	/* Sex */
	if (sex_on == TRUE)
	    if (!strcmp (item, "SX"))
	    {
		ni->sex = sstrdup (strtok (NULL, "\n"));
		continue;
	    }

	/* Location */
	if (location_on == TRUE)
	    if (!strcmp (item, "LO"))
	    {
		ni->location = sstrdup (strtok (NULL, "\n"));
		continue;
	    }

	/* Access List */
	if (!strcmp (item, "AC"))
	{
	    ++ni->accesscnt;
	    ni->access = srealloc (ni->access, sizeof (char *) *ni->accesscnt);
	    ni->access[ni->accesscnt - 1] = sstrdup (strtok (NULL, "\n"));
	    continue;
	}

	/* Chans List */
	if (!strcmp (item, "CH"))
	{
	    ++ni->chancnt;
	    ni->chans = srealloc (ni->chans, sizeof (char *) *ni->chancnt);
	    ni->chans[ni->chancnt - 1] = sstrdup (strtok (NULL, "\n"));
	    continue;
	}

	/* Successor List */
	if (!strcmp (item, "SU"))
	{
	    ++ni->successorcnt;
	    ni->successor = srealloc (ni->successor, sizeof (char *) *ni->successorcnt);
	    ni->successor[ni->successorcnt - 1] = sstrdup (strtok (NULL, "\n"));
	    continue;
	}

	/* Links List */
	if (!strcmp (item, "LN"))
	{
	    NickInfo **list, *lni;
	    char *tmp;

	    /* Ok. We do two things here. First, we add the nick to the
	       links list of the host nick. This saves a lot of cpu when
	       we want to know what nicks are linked to it. Second, we
	       'register' this nick and set its host, so everything else
	       works.
	     */
	    tmp = strtok (NULL, "\n");

	    ++ni->linkcnt;
	    ni->links = srealloc (ni->links, sizeof (char *) *ni->linkcnt);
	    ni->links[ni->linkcnt - 1] = sstrdup (tmp);

	    lni = scalloc (sizeof (NickInfo), 1);

	    strscpy (lni->nick, tmp, NICKLEN);
	    lni->host = sstrdup (ni->nick);
	    lni->flags = NF_LINKED;

	    list = &nicklist[NSHASH(lni->nick)];
	    lni->next = *list;
	    if (*list)
		(*list)->prev = lni;
	    *list = lni;

	    lin++;

	    continue;
	}

	/* Memo List */
	if (!strcmp (item, "MO"))
	{
	    mi = &ni->memos;

	    mi->memocnt++;
	    mi->memos = srealloc (mi->memos, sizeof (Memo) * mi->memocnt);
	    m = &mi->memos[mi->memocnt - 1];

	    m->number = atoi (strtok (NULL, " "));
	    m->flags = atoi (strtok (NULL, " "));
	    m->time = atoi (strtok (NULL, " "));
	    strscpy (m->sender, strtok (NULL, " "), NICKLEN);
	    m->text = sstrdup (strtok (NULL, "\n"));

	    min++;

	    continue;
	}

	/* Memo forward nick */
	if (!strcmp (item, "FW"))
	{
	    ni->forward = sstrdup (strtok (NULL, "\n"));
	    continue;
	}
    }

    fclose (f);

#ifdef HAVE_GETTIMEOFDAY
    if (!log)
    {
	gettimeofday (&now, NULL);
	timersub (&now, &start, &tmp);
	log_sameline (0, "finished in %d ms. (%d nicknames (%d links), %d memos)", tv2ms (&tmp), nin, lin, min);
    }
#endif
}

/* Write the NickServ db to disk. */
void save_ns_dbase ()
{
    FILE *f;
    NickInfo *ni;
    MemoInfo *mi;
    ChanInfo *ci;
    Memo *m;
    int j, log = 0, nout = 0, lout = 0, mout = 0;
    char backup[2048], **access, **links, **chans, **successor;
#ifdef HAVE_GETTIMEOFDAY
    struct timeval start, now, tmp;
#endif

    if (logupdates_on == TRUE)
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (1, "NS:DB: Saving... ");
#else
	log ("NS:DB: Saving... ");
#endif

    strcpy (backup, NICKSERV_DB);
    strcat (backup, ".save");
    remove (backup);

    if (rename (NICKSERV_DB, backup) < 0)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "Can't back up %s!", NICKSERV_DB);
#else
	log ("NS:DB: Can't back up %s!", NICKSERV_DB);
#endif
	log = 1;
    }

    f = fopen (NICKSERV_DB, "w");

    if (!f)
    {
#ifdef HAVE_GETTIMEOFDAY
	log_sameline (0, "Can't write to %s! ", NICKSERV_DB);
#else
	log ("NS:DB: Can't write to %s!", NICKSERV_DB);
#endif
	log = 1;

	if (rename (backup, NICKSERV_DB) < 0)
	    fatal ("NS:DB: Can't restore backup of %s!", NICKSERV_DB);
	return;
    }

#ifdef HAVE_GETTIMEOFDAY
    gettimeofday (&(start), NULL);
#endif

    /* Write the sstamp. */
    fprintf (f, "SS %d\n", sstamp);

    /* Write each nicknames info */
    for (ni = firstni (); ni; ni = nextni ())
    {
	/* Find their memos here. This lets us write memo things to NickInfo. */
	mi = &ni->memos;

	/* NickInfo - Only non linked nicks */
	if (!(ni->flags & NF_LINKED))
	{
	    fprintf (f, "NI %s %s %lu %lu %lu %d %d %d %lu %s %s\n",
		ni->nick, ni->pass, ni->registered, ni->lastseen,
		ni->timestamp, ni->sstamp, ni->flags, mi->memolimit,
		ni->key, ni->usermask, ni->real);

	    nout++;
	}

	/* ACCESS list */
	for (access = ni->access, j = 0; j < ni->accesscnt; access++, j++)
	    fprintf (f, "AC %s\n", *access);

	/* LINKS list */
	for (links = ni->links, j = 0; j < ni->linkcnt; links++, j++)
	{
	    fprintf (f, "LN %s\n", *links);
	    lout++;
	}

	/* CHANS list */
	for (chans = ni->chans, j = 0; j < ni->chancnt; chans++, j++)
	{
	    ci = cs_findchan (*chans);

	    if (!ci)
	    {
		/* This channel isn't registered anymore. Take it out
		   of the chans list, and continue on.
		 */
		free (*chans);

		--ni->chancnt;

		if (j < ni->chancnt)
		    bcopy (chans + 1, chans, (ni->chancnt - j) * sizeof (char *));

		if (ni->chancnt)
		    ni->chans = srealloc (ni->chans, ni->chancnt * sizeof (char *));
		else
		{
		    if (ni->chans)
			free (ni->chans);

		    ni->chans = NULL;
		}

		continue;
	    }

	    fprintf (f, "CH %s\n", *chans);
	}

	/* SUCCESSOR list */
	for (successor = ni->successor, j = 0; j < ni->successorcnt; successor++, j++)
	    fprintf (f, "SU %s\n", *successor);

	/* Memos */
	if (mi->memocnt)
	{
	    for (j = 0; j < mi->memocnt; j++)
	    {
		m = &mi->memos[j];

		fprintf (f, "MO %d %d %lu %s %s\n", m->number, m->flags,
		    m->time, m->sender, m->text);

		mout++;
	    }
	}

	/* Memo forward nick */
	if (ni->forward)
	{
	    /* Make sure it's still registered. */
	    if (findnick (ni->forward))
		fprintf (f, "FW %s\n", ni->forward);
	    else
	    {
		free (ni->forward);
		ni->forward = NULL;
	    }
	}

	/* Freeze reason */
	if (ni->freezereason)
	    fprintf (f, "FR %s\n", ni->freezereason);

	/* E-Mail */
	if (ni->email)
	{
	    if (email_on == TRUE)
		fprintf (f, "EM %s\n", ni->email);
	    else
		free (ni->email);
	}

	/* Temporary E-Mail */
	if (ni->temp)
	{
	    if (email_on == TRUE)
		fprintf (f, "TE %s\n", ni->temp);
	    else
		free (ni->temp);
	}

	/* URL */
	if (ni->url)
	{
	    if (url_on == TRUE)
		fprintf (f, "UR %s\n", ni->url);
	    else
		free (ni->url);
	}

	/* UIN */
	if (ni->uin)
	{
	    if (uin_on == TRUE)
		fprintf (f, "UN %d\n", ni->uin);
	    else
		ni->uin = 0;
	}

	/* Real Name */
	if (ni->name)
	{
	    if (name_on == TRUE)
		fprintf (f, "NA %s\n", ni->name);
	    else
		free (ni->name);
	}

	/* Age */
	if (ni->age)
	{
	    if (age_on == TRUE)
		fprintf (f, "AG %d\n", ni->age);
	    else
		ni->age = 0;
	}

	/* Sex */
	if (ni->sex)
	{
	    if (sex_on == TRUE)
		fprintf (f, "SX %s\n", ni->sex);
	    else
		free (ni->sex);
	}

	/* Location */
	if (ni->location)
	{
	    if (location_on == TRUE)
		fprintf (f, "LO %s\n", ni->location);
	    else
		free (ni->location);
	}
    }

    fclose (f);
    remove (backup);

#ifdef HAVE_GETTIMEOFDAY
    if (logupdates_on == TRUE && !log)
    {
	gettimeofday (&now, NULL);
	timersub (&now, &start, &tmp);
	log_sameline (0, "finished in %d ms. (%d nicknames (%d links), %d memos)", tv2ms (&tmp), nout, lout, mout);
    }
#endif
}

/* Return 1 if the first nickname is linked to the specified one, 0 otherwise. */
int islinked (const char *nick, NickInfo *ni)
{
    NickInfo *lni = findnick (nick);

    if (lni->host && !stricmp (lni->host, ni->nick))
	return 1;

    return 0;
}
